from unique_swot.services.report.config import RendererType, ReportRendererConfig
from unique_swot.services.report.delivery import ReportDeliveryService
from unique_swot.services.report.docx import convert_markdown_to_docx

__all__ = [
    "ReportRendererConfig",
    "RendererType",
    "convert_markdown_to_docx",
    "ReportDeliveryService",
]



from logging import getLogger

from jinja2 import Template
from unique_toolkit import LanguageModelService
from unique_toolkit._common.validators import LMI
from unique_toolkit.content import Content

from unique_swot.services.orchestrator.service import StepNotifier
from unique_swot.services.source_management.selection.config import (
    SourceSelectionConfig,
)
from unique_swot.services.source_management.selection.schema import (
    SourceSelectionResult,
)
from unique_swot.utils import generate_structured_output, get_content_chunk_title

_LOGGER = getLogger(__name__)


class SourceSelectionAgent:
    def __init__(
        self,
        *,
        llm_service: LanguageModelService,
        llm: LMI,
        source_selection_config: SourceSelectionConfig,
    ):
        self._llm_service = llm_service
        self._llm = llm
        self._source_selection_config = source_selection_config

    async def select(
        self, *, company_name: str, content: Content, step_notifier: StepNotifier
    ) -> SourceSelectionResult:
        _LOGGER.info(f"Selecting sources for {company_name}")
        system_prompt = self._compose_system_prompt(company_name=company_name)
        user_prompt = self._compose_user_prompt(
            company_name=company_name, content=content
        )
        document_title = get_content_chunk_title(content)
        notification_title = f"Reviewing `{document_title}`"

        await step_notifier.notify(
            title=notification_title,
        )

        response = await generate_structured_output(
            system_prompt=system_prompt,
            user_message=user_prompt,
            llm=self._llm,
            llm_service=self._llm_service,
            output_model=SourceSelectionResult,
        )

        if response is None:
            _LOGGER.error(f"Failed to select the source for {company_name}")
            response = SourceSelectionResult(
                should_select=True,
                reason="An error occured while selecting the source.",
                notification_message="An error occured while selecting the source. The source will be still considered for the SWOT as a safety measure.",
            )
        await step_notifier.notify(
            title=notification_title,
            description=response.notification_message,
            completed=True,
            progress=100,
        )
        return response

    def _compose_system_prompt(self, *, company_name: str) -> str:
        template = Template(self._source_selection_config.prompt_config.system_prompt)
        return template.render(company_name=company_name)

    def _compose_user_prompt(self, *, company_name: str, content: Content) -> str:
        selected_chunks = content.chunks[
            : self._source_selection_config.max_number_of_selected_chunks
        ]
        selected_chunks_texts = [chunk.text for chunk in selected_chunks]

        template = Template(self._source_selection_config.prompt_config.user_prompt)
        return template.render(
            company_name=company_name, selected_chunks=selected_chunks_texts
        )



"""Tests for DOCX conversion utilities."""

from unittest.mock import Mock

from unique_swot.services.report.docx import (
    convert_markdown_to_docx,
)


def test_convert_markdown_to_docx_success():
    """Test successful markdown to DOCX conversion."""
    markdown = "# Test Report\n\nContent"
    fields = {"title": "SWOT Analysis", "date": "2024-01-01"}

    docx_generator = Mock()
    docx_generator.parse_markdown_to_list_content_fields.return_value = [
        {"type": "heading", "content": "Test Report"},
        {"type": "paragraph", "content": "Content"},
    ]
    docx_generator.generate_from_template.return_value = b"fake docx bytes"

    result = convert_markdown_to_docx(markdown, docx_generator, fields)

    assert result == b"fake docx bytes"
    docx_generator.parse_markdown_to_list_content_fields.assert_called_once_with(
        markdown
    )
    docx_generator.generate_from_template.assert_called_once()


def test_convert_markdown_to_docx_with_complex_markdown():
    """Test conversion with complex markdown elements."""
    markdown = """
# Main Title

## Section 1

This is a paragraph with **bold** and *italic* text.

- Bullet point 1
- Bullet point 2

## Section 2

1. Numbered item 1
2. Numbered item 2

> Quote block

[Link text](https://example.com)
"""
    fields = {"title": "Report", "date": "2024-01-01"}

    docx_generator = Mock()
    docx_generator.parse_markdown_to_list_content_fields.return_value = [
        {"type": "heading", "level": 1, "content": "Main Title"},
        {"type": "heading", "level": 2, "content": "Section 1"},
        {"type": "paragraph", "content": "This is a paragraph..."},
    ]
    docx_generator.generate_from_template.return_value = b"docx content"

    result = convert_markdown_to_docx(markdown, docx_generator, fields)

    assert result == b"docx content"
    # Verify markdown was parsed
    docx_generator.parse_markdown_to_list_content_fields.assert_called_once()


def test_convert_markdown_to_docx_passes_fields():
    """Test that template fields are passed correctly."""
    markdown = "# Report"
    fields = {
        "title": "ACME Corp SWOT Analysis",
        "date": "2024-01-15",
        "author": "Analysis Team",
    }

    docx_generator = Mock()
    docx_generator.parse_markdown_to_list_content_fields.return_value = []
    docx_generator.generate_from_template.return_value = b"docx"

    convert_markdown_to_docx(markdown, docx_generator, fields)

    # Verify fields were passed to generate_from_template
    call_args = docx_generator.generate_from_template.call_args
    assert call_args[0][1] == fields


def test_convert_markdown_to_docx_returns_none_on_failure():
    """Test that conversion returns None on failure."""
    markdown = "# Report"
    fields = {"title": "Report"}

    docx_generator = Mock()
    docx_generator.parse_markdown_to_list_content_fields.return_value = []
    docx_generator.generate_from_template.return_value = None  # Failure

    result = convert_markdown_to_docx(markdown, docx_generator, fields)

    assert result is None


def test_convert_markdown_to_docx_empty_markdown():
    """Test conversion with empty markdown."""
    markdown = ""
    fields = {"title": "Report"}

    docx_generator = Mock()
    docx_generator.parse_markdown_to_list_content_fields.return_value = []
    docx_generator.generate_from_template.return_value = b"empty docx"

    result = convert_markdown_to_docx(markdown, docx_generator, fields)

    assert result == b"empty docx"



import json
from collections import defaultdict

from unique_swot.services.generation.context import SWOTComponent
from unique_swot.services.generation.models.base import SWOTReportComponentSection
from unique_swot.utils import generate_unique_id


class SWOTReportRegistry:
    def __init__(self):
        self._by_id: dict[str, SWOTReportComponentSection] = {}
        self._id_to_component: dict[str, SWOTComponent] = {}
        self._by_component: dict[
            SWOTComponent, dict[str, SWOTReportComponentSection]
        ] = defaultdict(dict)

    def register_section(
        self, component: SWOTComponent, section: SWOTReportComponentSection
    ) -> str:
        section_id = generate_unique_id(f"section_{component.value}")
        self._by_id[section_id] = section
        self._id_to_component[section_id] = component
        self._by_component[component][section_id] = section
        return section_id

    def retrieve_section(self, id: str) -> SWOTReportComponentSection | None:
        return self._by_id.get(id)

    def retrieve_sections_for_component(
        self, component: SWOTComponent, exclude_items: bool = True
    ) -> str:
        exclude_fields = {"items"} if exclude_items else None
        sections = self._by_component.get(component, {})
        return json.dumps(
            {
                id: section.model_dump(exclude=exclude_fields)
                for id, section in sections.items()
            },
            indent=1,
        )

    def update_section(self, id: str, section: SWOTReportComponentSection) -> None:
        target_component = self._id_to_component.get(id)
        if target_component is None:
            raise KeyError(f"Section id {id} not found in registry")
        self._by_id[id] = section
        self._by_component[target_component][id] = section

    def retrieve_component_sections(
        self, component: SWOTComponent
    ) -> list[SWOTReportComponentSection]:
        return list(self._by_component.get(component, {}).values())



"""Tests for the SourceCollectionManager."""

from datetime import datetime
from unittest.mock import AsyncMock, Mock, patch

import pytest
from unique_toolkit.content import Content, ContentChunk

from unique_swot.services.session.schema import UniqueCompanyListing
from unique_swot.services.source_management.collection.base import (
    CollectionContext,
    SourceCollectionManager,
)


def _make_company():
    """Helper to create test company listing."""
    params = {
        "sourceRef": 123.0,
        "name": "ACME Corp",
        "displayName": "ACME Corporation",
        "country": "US",
        "tickers": [],
        "sourceUrl": "https://example.com/acme",
        "source": "test",
    }
    return UniqueCompanyListing.model_validate(params)


def _make_content(content_id="content_1", title="Test Doc"):
    """Helper to create test Content."""
    return Content(
        id=content_id,
        title=title,
        key=f"{title}.pdf",
        chunks=[
            ContentChunk(
                id=content_id,
                chunk_id="chunk_1",
                title=title,
                key=f"{title}.pdf",
                text="Test content",
                start_page=1,
                end_page=1,
                order=0,
            )
        ],
    )


@pytest.mark.asyncio
async def test_collection_manager_collects_all_sources(
    mock_knowledge_base_service,
    mock_quartr_service,
    mock_docx_generator,
    mock_step_notifier,
):
    """Test that collection manager collects from all enabled sources."""
    context = CollectionContext(
        company=_make_company(),
        use_earnings_calls=True,
        upload_scope_id_earnings_calls="scope_123",
        earnings_call_start_date=datetime(2023, 1, 1),
        use_web_sources=True,
        metadata_filter={"key": "value"},
    )

    # Mock the collection functions
    with (
        patch(
            "unique_swot.services.source_management.collection.base.collect_knowledge_base",
            new_callable=AsyncMock,
            return_value=[_make_content("kb_1", "KB Doc")],
        ),
        patch(
            "unique_swot.services.source_management.collection.base.collect_earnings_calls",
            new_callable=AsyncMock,
            return_value=[_make_content("ec_1", "Earnings Call")],
        ),
        patch(
            "unique_swot.services.source_management.collection.base.collect_web_sources",
            return_value=[_make_content("web_1", "Web Source")],
        ),
    ):
        manager = SourceCollectionManager(
            context=context,
            knowledge_base_service=mock_knowledge_base_service,
            quartr_service=mock_quartr_service,
            earnings_call_docx_generator_service=mock_docx_generator,
        )

        sources = await manager.collect(step_notifier=mock_step_notifier)

        assert len(sources) == 3
        assert any(s.title == "KB Doc" for s in sources)
        assert any(s.title == "Earnings Call" for s in sources)
        assert any(s.title == "Web Source" for s in sources)


@pytest.mark.asyncio
async def test_collection_manager_skips_disabled_sources(
    mock_knowledge_base_service, mock_step_notifier
):
    """Test that disabled sources are skipped."""
    context = CollectionContext(
        company=_make_company(),
        use_earnings_calls=False,
        upload_scope_id_earnings_calls="scope_123",
        earnings_call_start_date=datetime(2023, 1, 1),
        use_web_sources=False,
        metadata_filter={"key": "value"},
    )

    with patch(
        "unique_swot.services.source_management.collection.base.collect_knowledge_base",
        new_callable=AsyncMock,
        return_value=[_make_content("kb_1", "KB Doc")],
    ):
        manager = SourceCollectionManager(
            context=context,
            knowledge_base_service=mock_knowledge_base_service,
            quartr_service=None,
            earnings_call_docx_generator_service=Mock(),
        )

        sources = await manager.collect(step_notifier=mock_step_notifier)

        # Only KB sources should be collected
        assert len(sources) == 1
        assert sources[0].title == "KB Doc"


@pytest.mark.asyncio
async def test_collection_manager_no_metadata_filter_skips_kb(
    mock_knowledge_base_service, mock_step_notifier
):
    """Test that no metadata filter skips KB collection."""
    context = CollectionContext(
        company=_make_company(),
        use_earnings_calls=False,
        upload_scope_id_earnings_calls="scope_123",
        earnings_call_start_date=datetime(2023, 1, 1),
        use_web_sources=False,
        metadata_filter=None,  # No filter
    )

    manager = SourceCollectionManager(
        context=context,
        knowledge_base_service=mock_knowledge_base_service,
        quartr_service=None,
        earnings_call_docx_generator_service=Mock(),
    )

    sources = await manager.collect(step_notifier=mock_step_notifier)

    # No sources should be collected
    assert len(sources) == 0


@pytest.mark.asyncio
async def test_collection_manager_no_quartr_service_skips_earnings(
    mock_knowledge_base_service, mock_step_notifier
):
    """Test that missing Quartr service skips earnings calls."""
    context = CollectionContext(
        company=_make_company(),
        use_earnings_calls=True,
        upload_scope_id_earnings_calls="scope_123",
        earnings_call_start_date=datetime(2023, 1, 1),
        use_web_sources=False,
        metadata_filter=None,
    )

    manager = SourceCollectionManager(
        context=context,
        knowledge_base_service=mock_knowledge_base_service,
        quartr_service=None,  # No Quartr service
        earnings_call_docx_generator_service=Mock(),
    )

    sources = await manager.collect(step_notifier=mock_step_notifier)

    # No sources should be collected
    assert len(sources) == 0


@pytest.mark.asyncio
async def test_collection_manager_sends_notifications(
    mock_knowledge_base_service, mock_step_notifier
):
    """Test that notifications are sent during collection."""
    context = CollectionContext(
        company=_make_company(),
        use_earnings_calls=False,
        upload_scope_id_earnings_calls="scope_123",
        earnings_call_start_date=datetime(2023, 1, 1),
        use_web_sources=False,
        metadata_filter={"key": "value"},
    )

    with patch(
        "unique_swot.services.source_management.collection.base.collect_knowledge_base",
        new_callable=AsyncMock,
        return_value=[_make_content()],
    ):
        manager = SourceCollectionManager(
            context=context,
            knowledge_base_service=mock_knowledge_base_service,
            quartr_service=None,
            earnings_call_docx_generator_service=Mock(),
        )

        await manager.collect(step_notifier=mock_step_notifier)

        # Verify notifications were sent
        assert mock_step_notifier.notify.await_count >= 2  # Start and end


@pytest.mark.asyncio
async def test_collection_manager_notification_title(
    mock_knowledge_base_service, mock_step_notifier
):
    """Test notification title property."""
    context = CollectionContext(
        company=_make_company(),
        use_earnings_calls=False,
        upload_scope_id_earnings_calls="scope_123",
        earnings_call_start_date=datetime(2023, 1, 1),
        use_web_sources=False,
        metadata_filter=None,
    )

    manager = SourceCollectionManager(
        context=context,
        knowledge_base_service=mock_knowledge_base_service,
        quartr_service=None,
        earnings_call_docx_generator_service=Mock(),
    )

    assert manager.notification_title == "**Collecting Sources**"


@pytest.mark.asyncio
async def test_collection_context_is_immutable():
    """Test that CollectionContext is frozen/immutable."""
    context = CollectionContext(
        company=_make_company(),
        use_earnings_calls=True,
        upload_scope_id_earnings_calls="scope_123",
        earnings_call_start_date=datetime(2023, 1, 1),
        use_web_sources=True,
        metadata_filter={"key": "value"},
    )

    # Attempting to modify should raise an error
    with pytest.raises(Exception):  # Pydantic raises ValidationError
        context.use_earnings_calls = False



from unique_toolkit import ChatService
from unique_toolkit.chat.schemas import MessageExecutionUpdateStatus

EXECUTION_OPTIONS = {"toolChoices": ["SWOT"]}


class ProgressNotifier:
    def __init__(self, chat_service: ChatService, company_name: str):
        self._chat_service = chat_service
        self._company_name = company_name
        self._message_execution = None
        self._title = ""
        self._step_size = 0
        self._progress = 0

    @property
    def title(self) -> str:
        return self._title

    @property
    def step_size(self) -> int | float:
        return self._step_size

    @step_size.setter
    def step_size(self, value: int | float):
        self._step_size = value

    @property
    def progress(self) -> int:
        return self._progress

    @progress.setter
    def progress(self, value: int | float):
        value = self._bound_and_convert_to_int(value)
        self._progress = value

    async def start(self, *, title: str = ""):
        self._title = title
        self._message_execution = (
            await self._chat_service.create_message_execution_async(
                message_id=self._chat_service.assistant_message_id,
                is_queueable=False,
                execution_options=EXECUTION_OPTIONS,
                progress_title=self._title,
                percentage_completed=0,
            )
        )

    async def increment(self, fraction: float):
        self.progress += fraction * self._step_size
        await self._chat_service.update_message_execution_async(
            message_id=self._chat_service.assistant_message_id,
            percentage_completed=self.progress,
            progress_title=self._title,
        )

    async def update(self, *, progress: int | float, title: str | None = None):
        self.progress = progress
        if title:
            self._title = title

        await self._chat_service.update_message_execution_async(
            message_id=self._chat_service.assistant_message_id,
            percentage_completed=self._bound_and_convert_to_int(progress),
            progress_title=self._title,
        )

    async def finish(self, failed: bool = False):
        if self._message_execution:
            await self._chat_service.update_message_execution_async(
                message_id=self._chat_service.assistant_message_id,
                percentage_completed=100,
                status=MessageExecutionUpdateStatus.COMPLETED
                if not failed
                else MessageExecutionUpdateStatus.FAILED,
            )

    @staticmethod
    def _bound_and_convert_to_int(progress: int | float) -> int:
        return int(max(0, min(progress, 100)))



"""
SWOT Analysis Data Schemas

This module defines the core data structures used throughout the SWOT analysis system.
These schemas represent the workflow from planning to execution, including data sources,
analysis steps, and results.
"""

from enum import StrEnum
from typing import Callable, Generic, TypeVar

from jinja2 import Template
from pydantic import BaseModel, Field

from unique_swot.services.generation.context import SWOTComponent


# Type definitions for SWOT operations
class SWOTOperation(StrEnum):
    GENERATE = "generate"
    MODIFY = "modify"
    NOT_REQUESTED = "not_requested"


class SWOTStepPlan(BaseModel):
    operation: SWOTOperation
    modify_instruction: str | None = Field(
        description="Custom instruction for the modify operation. This is only used if the operation is modify."
    )


class SWOTStepResult(SWOTStepPlan):
    result: str


TStep = TypeVar("TStep", bound=SWOTStepPlan)


class SWOT(BaseModel, Generic[TStep]):
    objective: str = Field(description="The objective of the plan to be executed")

    strengths: TStep = Field(description="The step to analyze the strengths")
    weaknesses: TStep = Field(description="The step to analyze the weaknesses")
    opportunities: TStep = Field(description="The step to analyze the opportunities")
    threats: TStep = Field(description="The step to analyze the threats")


class SWOTPlan(SWOT[SWOTStepPlan]):
    # from unique_swot.services.generation import SWOTComponent

    def validate_swot_plan(self) -> None:
        for step in [self.strengths, self.weaknesses, self.opportunities, self.threats]:
            if (
                step.operation == SWOTOperation.MODIFY
                and step.modify_instruction is None
            ):
                raise ValueError("Modify instruction is required for modify operations")

    def get_step_result(self, component: SWOTComponent) -> SWOTStepPlan:
        # from unique_swot.services.generation import SWOTComponent

        match component:
            case SWOTComponent.STRENGTHS:
                return self.strengths
            case SWOTComponent.WEAKNESSES:
                return self.weaknesses
            case SWOTComponent.OPPORTUNITIES:
                return self.opportunities
            case SWOTComponent.THREATS:
                return self.threats
            case _:
                raise ValueError(f"Invalid component: {component}")

    def __len__(self) -> int:
        return len(
            [
                step
                for step in [
                    self.strengths,
                    self.weaknesses,
                    self.opportunities,
                    self.threats,
                ]
                if step.operation != SWOTOperation.NOT_REQUESTED
            ]
        )

    def convert_plan_to_list_of_steps(self) -> list[tuple[SWOTComponent, SWOTStepPlan]]:
        return [
            (SWOTComponent.STRENGTHS, self.strengths),
            (SWOTComponent.WEAKNESSES, self.weaknesses),
            (SWOTComponent.OPPORTUNITIES, self.opportunities),
            (SWOTComponent.THREATS, self.threats),
        ]


class SWOTResult(SWOT[SWOTStepResult]):
    # from unique_swot.services.generation import SWOTComponent

    @classmethod
    def init_from_plan(cls, *, plan: SWOTPlan) -> "SWOTResult":
        strengths_result = SWOTStepResult(
            operation=plan.strengths.operation,
            modify_instruction=plan.strengths.modify_instruction,
            result="",
        )
        weaknesses_result = SWOTStepResult(
            operation=plan.weaknesses.operation,
            modify_instruction=plan.weaknesses.modify_instruction,
            result="",
        )
        opportunities_result = SWOTStepResult(
            operation=plan.opportunities.operation,
            modify_instruction=plan.opportunities.modify_instruction,
            result="",
        )
        threats_result = SWOTStepResult(
            operation=plan.threats.operation,
            modify_instruction=plan.threats.modify_instruction,
            result="",
        )
        return cls(
            objective=plan.objective,
            strengths=strengths_result,
            weaknesses=weaknesses_result,
            opportunities=opportunities_result,
            threats=threats_result,
        )

    def assign_result(self, component: SWOTComponent, result: str) -> None:
        # from unique_swot.services.generation import SWOTComponent

        match component:
            case SWOTComponent.STRENGTHS:
                self.strengths.result = result
            case SWOTComponent.WEAKNESSES:
                self.weaknesses.result = result
            case SWOTComponent.OPPORTUNITIES:
                self.opportunities.result = result
            case SWOTComponent.THREATS:
                self.threats.result = result
            case _:
                raise ValueError(f"Invalid component: {component}")

    def to_markdown_report(
        self, markdown_jinja_template: str, processor: Callable[[str], str]
    ) -> str:
        report = Template(markdown_jinja_template).render(**self.model_dump())
        return processor(report)



from pydantic import Field

from unique_swot.utils import StructuredOutputWithNotification


class SourceSelectionResult(StructuredOutputWithNotification):
    """This class is responsible for the result of the source selection."""

    reason: str = Field(description="The reason for the source selection")
    should_select: bool = Field(description="Whether the source should be selected")



from logging import getLogger
from typing import Callable

from jinja2 import Template
from unique_toolkit import ChatService
from unique_toolkit._common.docx_generator import DocxGeneratorService

from unique_swot.services.citations import CitationManager
from unique_swot.services.generation.models.base import SWOTReportComponents
from unique_swot.services.report.config import RendererType, ReportRendererConfig
from unique_swot.services.report.docx import (
    convert_markdown_to_docx,
)
from unique_swot.services.session import SwotAnalysisSessionConfig
from unique_swot.utils import convert_content_chunk_to_reference

_LOGGER = getLogger(__name__)


class ReportDeliveryService:
    """
    Handles the delivery of SWOT analysis reports to the chat.

    This service is responsible for:
    - Converting markdown reports to appropriate formats (DOCX/Markdown)
    - Uploading content to chat
    - Creating content references
    - Modifying assistant messages with the delivered content
    """

    def __init__(
        self,
        *,
        chat_service: ChatService,
        docx_renderer: DocxGeneratorService,
        citation_manager: CitationManager,
        renderer_config: ReportRendererConfig,
    ):
        self._chat_service = chat_service
        self._docx_renderer = docx_renderer
        self._citation_manager = citation_manager
        self._renderer_config = renderer_config

    def render_report(
        self,
        citation_fn: Callable[[str], str],
        executive_summary: str | None = None,
        body: str | None = None,
    ) -> str:
        # Render using the template
        markdown_report = Template(
            self._renderer_config.report_structure_template
        ).render(
            executive_summary=executive_summary,
            body=body,
        )
        return citation_fn(markdown_report)

    def render_body(self, result: SWOTReportComponents) -> str:
        return Template(self._renderer_config.report_body_template).render(
            **result.model_dump(),
        )

    def deliver_report(
        self,
        *,
        executive_summary: str,
        body: str,
        session_config: SwotAnalysisSessionConfig,
        docx_template_fields: dict[str, str],
        ingest_docx: bool,
    ) -> str:
        """
        Delivers a SWOT analysis report to the chat.

        Args:
            session_config: The session configuration containing company info
            result: List of consolidated SWOT reports (one per component)
            docx_template_fields: The fields to be used in the DOCX template
            ingest_docx: Whether to ingest the DOCX file
        """

        renderer_type = self._renderer_config.renderer_type

        executive_summary, body = map(
            lambda content: self._citation_manager.map_citations_to_report(
                content, self._renderer_config.renderer_type
            ),
            [executive_summary, body],
        )

        number_of_citations = len(self._citation_manager.get_citations_for_docx())

        full_report = Template(self._renderer_config.report_structure_template).render(
            executive_summary=executive_summary,
            body=body,
            citations=self._citation_manager.get_citations_for_docx()
            if renderer_type == RendererType.DOCX
            else None,
        )

        match renderer_type:
            case RendererType.DOCX:
                self._deliver_docx_report(
                    session_config=session_config,
                    markdown_report=full_report,
                    template_fields=docx_template_fields,
                    ingest_docx=ingest_docx,
                    sequence_number=number_of_citations,
                )
            case RendererType.CHAT:
                self._deliver_markdown_report(
                    markdown_report=full_report,
                )
            case _:
                raise ValueError(f"Invalid renderer type: {renderer_type}")

        return full_report

    def _convert_consolidated_reports_to_markdown(
        self,
        consolidated_reports: SWOTReportComponents,
        markdown_jinja_template: str,
        processor: Callable[[str], str],
    ) -> str:
        """
        Converts a list of consolidated reports into a markdown report.

        Args:
            consolidated_reports: List of consolidated reports (one per component)
            markdown_jinja_template: The Jinja2 template for rendering
            processor: Function to process the markdown (e.g., add citations)

        Returns:
            Formatted markdown report
        """

        # Render using the template
        markdown_report = Template(markdown_jinja_template).render(
            **consolidated_reports.model_dump()
        )
        return processor(markdown_report)

    def _deliver_docx_report(
        self,
        *,
        markdown_report: str,
        template_fields: dict[str, str],
        session_config: SwotAnalysisSessionConfig,
        ingest_docx: bool,
        sequence_number: int,
    ) -> None:
        """Converts markdown to DOCX and delivers it as an attachment"""

        # Convert markdown to DOCX
        docx_bytes = convert_markdown_to_docx(
            markdown_report, self._docx_renderer, template_fields
        )
        if docx_bytes is None:
            raise ValueError("Failed to convert markdown to DOCX")

        # Upload to chat
        document_name = (
            f"{session_config.company_listing.name} SWOT Analysis Report.docx"
        )
        content = self._chat_service.upload_to_chat_from_bytes(
            content=docx_bytes,
            content_name=document_name,
            mime_type="application/vnd.openxmlformats-officedocument.wordprocessingml.document",
            skip_ingestion=not ingest_docx,
        )

        # Create content reference
        content_reference = convert_content_chunk_to_reference(
            content_or_chunk=content,
            sequence_number=sequence_number,
        )

        # Modify assistant message
        self._chat_service.modify_assistant_message(
            message_id=self._chat_service.assistant_message_id,
            content=f"Here is the full Swot Analysis report for {session_config.company_listing.name} in DOCX format <sup>{sequence_number}</sup>.",
            references=[content_reference],
        )

        _LOGGER.info(
            f"Successfully delivered DOCX report for message {self._chat_service.assistant_message_id}"
        )

    def _deliver_markdown_report(
        self,
        *,
        markdown_report: str,
    ) -> None:
        """Delivers the markdown report directly to the chat"""
        references = self._citation_manager.get_references(
            message_id=self._chat_service.assistant_message_id
        )

        self._chat_service.modify_assistant_message(
            message_id=self._chat_service.assistant_message_id,
            content=markdown_report,
            references=references,
        )

        _LOGGER.info(
            f"Successfully delivered markdown report for message {self._chat_service.assistant_message_id}"
        )



class SWOTReportAgentException(Exception):
    pass


class FailedToGeneratePlanException(SWOTReportAgentException):
    pass


class InvalidPlanException(SWOTReportAgentException):
    pass


class FailedToExtractFactsException(SWOTReportAgentException):
    pass


class FailedToCreateNewSectionException(SWOTReportAgentException):
    pass


class FailedToUpdateExistingSectionException(SWOTReportAgentException):
    pass


class SectionNotFoundException(SWOTReportAgentException):
    pass


class SourceFactsIdsNotFoundException(SWOTReportAgentException):
    pass


class InvalidCommandException(SWOTReportAgentException):
    pass



import json
from logging import getLogger
from typing import Any, Sequence

from jinja2 import Template
from unique_toolkit import LanguageModelService
from unique_toolkit._common.validators import LMI

from unique_swot.services.generation.agentic.commands import (
    create_new_section,
    update_existing_section,
)
from unique_swot.services.generation.agentic.config import AgenticGeneratorConfig
from unique_swot.services.generation.agentic.exceptions import (
    FailedToExtractFactsException,
    FailedToGeneratePlanException,
    InvalidCommandException,
)
from unique_swot.services.generation.agentic.executor import AgenticPlanExecutor
from unique_swot.services.generation.agentic.prompts.commands.config import (
    CommandsPromptConfig,
)
from unique_swot.services.generation.agentic.prompts.definition.config import (
    ComponentDefinitionPromptConfig,
    get_component_definition,
)
from unique_swot.services.generation.agentic.prompts.extraction.config import (
    ExtractionPromptConfig,
)
from unique_swot.services.generation.agentic.prompts.plan.config import PlanPromptConfig
from unique_swot.services.generation.context import SWOTComponent
from unique_swot.services.generation.models.base import SWOTExtractionFactsList
from unique_swot.services.generation.models.plan import (
    GenerationPlan,
    GenerationPlanCommandType,
)
from unique_swot.services.generation.models.registry import SWOTReportRegistry
from unique_swot.services.orchestrator.service import StepNotifier
from unique_swot.utils import generate_structured_output, generate_unique_id

_LOGGER = getLogger(__name__)


async def handle_generate_operation(
    *,
    component: SWOTComponent,
    source_batches: list[dict[str, str]],
    step_notifier: StepNotifier,
    company_name: str,
    llm: LMI,
    llm_service: LanguageModelService,
    notification_title: str,
    swot_report_registry: SWOTReportRegistry,
    executor: AgenticPlanExecutor,
    config: AgenticGeneratorConfig,
) -> None:
    try:
        # Extract the items for the component
        extracted_facts = await _extract_facts(
            company_name=company_name,
            component=component,
            source_batches=source_batches,
            step_notifier=step_notifier,
            llm=llm,
            llm_service=llm_service,
            notification_title=notification_title,
            prompts_config=config.prompts_config.extraction_prompt_config,
            component_definition_prompt_config=config.prompts_config.definition_prompt_config,
        )

        # Skip if list of facts is empty
        if len(extracted_facts) == 0:
            _LOGGER.warning(
                f"No facts extracted for component {component}. Skipping..."
            )
            return

        fact_id_map = {generate_unique_id("fact_"): fact for fact in extracted_facts}

        # Plan the generation for the component
        plan = await _generate_plan(
            component=component,
            fact_id_map=fact_id_map,
            step_notifier=step_notifier,
            company_name=company_name,
            llm=llm,
            llm_service=llm_service,
            notification_title=notification_title,
            swot_report_registry=swot_report_registry,
            prompts_config=config.prompts_config.plan_prompt_config,
        )

        # Execute the generation for the component
        results = await _execute_plan(
            plan=plan,
            component=component,
            fact_id_map=fact_id_map,
            swot_report_registry=swot_report_registry,
            llm=llm,
            llm_service=llm_service,
            company_name=company_name,
            executor=executor,
            prompts_config=config.prompts_config.commands_prompt_config,
        )
        # Register or update the sections in the registry
        _handle_execution_results(
            plan=plan,
            results=results,
            component=component,
            swot_report_registry=swot_report_registry,
        )

    except FailedToExtractFactsException as e:
        _LOGGER.exception(
            f"Failed to extract facts for component {component}", exc_info=e
        )
        await step_notifier.notify(
            title=notification_title,
            description=f"An error occured while extracting facts for component {component}. This batch will be skipped.",
        )
    except FailedToGeneratePlanException as e:
        _LOGGER.exception(
            f"Failed to generate plan for component {component}", exc_info=e
        )
        await step_notifier.notify(
            title=notification_title,
            description=f"An error occured while generating plan to update the SWOT report for component {component}. Extracted information will be discarded.",
        )
    except Exception as e:
        _LOGGER.exception(
            f"Failed to generate operation for component {component}", exc_info=e
        )
        await step_notifier.notify(
            title=notification_title,
            description=f"An unexpected error occured while updating the SWOT report for component {component}. Skipping this batch.",
        )


async def _extract_facts(
    *,
    company_name: str,
    component: SWOTComponent,
    source_batches: list[dict[str, str]],
    step_notifier: StepNotifier,
    llm: LMI,
    llm_service: LanguageModelService,
    notification_title: str,
    prompts_config: ExtractionPromptConfig,
    component_definition_prompt_config: ComponentDefinitionPromptConfig,
) -> list[str]:
    component_definition = get_component_definition(
        component, component_definition_prompt_config
    )
    system_prompt = Template(prompts_config.system_prompt).render(
        company_name=company_name,
        component=component,
        component_definition=component_definition,
    )

    user_message = Template(prompts_config.user_prompt).render(
        company_name=company_name,
        component=component,
        source_batches=source_batches,
    )

    extracted_facts = await generate_structured_output(
        system_prompt=system_prompt,
        user_message=user_message,
        llm_service=llm_service,
        llm=llm,
        output_model=SWOTExtractionFactsList,
    )

    if extracted_facts is None:
        _LOGGER.error(f"Failed to extract facts for component {component}")
        await step_notifier.notify(
            title=notification_title,
            description=f"Failed to extract facts for component {component} from this batch.",
        )
        raise FailedToExtractFactsException(
            f"No facts extracted for component {component}"
        )

    return extracted_facts.facts


async def _generate_plan(
    *,
    component: SWOTComponent,
    fact_id_map: dict[str, str],
    step_notifier: StepNotifier,
    company_name: str,
    llm: LMI,
    llm_service: LanguageModelService,
    notification_title: str,
    swot_report_registry: SWOTReportRegistry,
    prompts_config: PlanPromptConfig,
) -> GenerationPlan:
    system_prompt = Template(prompts_config.system_prompt).render(
        company_name=company_name,
        component=component,
    )

    existing_sections_view = swot_report_registry.retrieve_sections_for_component(
        component=component,
        exclude_items=False,
    )

    fact_view = json.dumps(fact_id_map, indent=1)

    user_message = Template(prompts_config.user_prompt).render(
        fact_view=fact_view,
        existing_sections_view=existing_sections_view,
    )

    plan = await generate_structured_output(
        system_prompt=system_prompt,
        user_message=user_message,
        llm_service=llm_service,
        llm=llm,
        output_model=GenerationPlan,
    )
    if plan is None:
        _LOGGER.error(f"Failed to generate plan for component {component}")
        raise FailedToGeneratePlanException(
            f"Failed to generate plan for component {component}"
        )

    _LOGGER.info(f"Generated plan for component {component}")
    await step_notifier.notify(
        title=notification_title,
        description=plan.notification_message,
    )
    return plan


async def _execute_plan(
    *,
    plan: GenerationPlan,
    component: SWOTComponent,
    fact_id_map: dict[str, str],
    swot_report_registry: SWOTReportRegistry,
    llm: LMI,
    llm_service: LanguageModelService,
    company_name: str,
    executor: AgenticPlanExecutor,
    prompts_config: CommandsPromptConfig,
) -> Sequence[Exception | Any]:
    for command in plan.commands:
        match command.command:
            case GenerationPlanCommandType.CREATE_SECTION:
                executor.add(
                    create_new_section,
                    llm=llm,
                    llm_service=llm_service,
                    company_name=company_name,
                    component=component,
                    instruction=command.instruction,
                    command=command,
                    fact_id_map=fact_id_map,
                    prompts_config=prompts_config.create_new_section_prompt_config,
                )

            case GenerationPlanCommandType.UPDATE_SECTION:
                executor.add(
                    update_existing_section,
                    llm=llm,
                    llm_service=llm_service,
                    company_name=company_name,
                    component=component,
                    instruction=command.instruction,
                    command=command,
                    fact_id_map=fact_id_map,
                    swot_report_registry=swot_report_registry,
                    prompts_config=prompts_config.update_existing_section_prompt_config,
                )
            case _:
                raise InvalidCommandException(f"Invalid command: {command.command}")

    results = await executor.run()

    return results


def _handle_execution_results(
    *,
    plan: GenerationPlan,
    results: Sequence[Exception | Any],
    component: SWOTComponent,
    swot_report_registry: SWOTReportRegistry,
) -> list[Exception]:
    exceptions = []
    for command, result in zip(plan.commands, results):
        if isinstance(result, Exception):
            _LOGGER.exception(
                f"Error executing command {command.command}", exc_info=result
            )
            exceptions.append(result)
        else:
            match command.command:
                case GenerationPlanCommandType.CREATE_SECTION:
                    swot_report_registry.register_section(
                        component=component, section=result
                    )
                case GenerationPlanCommandType.UPDATE_SECTION:
                    swot_report_registry.update_section(
                        id=command.target_section_id,  # type: ignore (if target id is None, the result is not a section)
                        section=result,
                    )
                case _:
                    raise InvalidCommandException(f"Invalid command: {command.command}")
    return exceptions



from logging import getLogger

from pydantic import BaseModel, Field
from unique_toolkit._common.validators import LMI, get_LMI_default_field
from unique_toolkit.agentic.tools.config import get_configuration_dict
from unique_toolkit.language_model.infos import (
    LanguageModelName,
)

from unique_swot.services.generation.agentic.config import AgenticGeneratorConfig

# from unique_swot.services.generation.extraction.config import (
#     ExtractionConfig,
# )
# from unique_swot.services.generation.reporting.config import ReportingConfig

_LOGGER = getLogger(__name__)

_DEFAULT_LANGUAGE_MODEL = LanguageModelName.AZURE_GPT_5_2025_0807
_DEFAULT_BATCH_SIZE = 30
_DEFAULT_MAX_TOKENS_PER_BATCH = 30_000


class ReportGenerationConfig(BaseModel):
    """
    Configuration settings for SWOT report generation.

    Controls the language model, batching behavior, and token limits for report generation.

    Attributes:
        language_model: The language model to use for generation
        batch_size: Number of sources to process in each batch
        max_tokens_per_batch: Maximum tokens allowed per batch to prevent overflow
    """

    model_config = get_configuration_dict()

    language_model: LMI = get_LMI_default_field(
        _DEFAULT_LANGUAGE_MODEL, description="The language model to use for generation"
    )
    agentic_generator_config: AgenticGeneratorConfig = Field(
        default_factory=AgenticGeneratorConfig,
        description="The configuration for the agentic generator.",
    )



import json
from logging import getLogger

from jinja2 import Template
from unique_toolkit import LanguageModelService
from unique_toolkit._common.validators import LMI

from unique_swot.services.generation.agentic.exceptions import (
    FailedToCreateNewSectionException,
    FailedToUpdateExistingSectionException,
    InvalidPlanException,
    SectionNotFoundException,
)
from unique_swot.services.generation.agentic.prompts.commands.create_new_section.config import (
    CreateNewSectionPromptConfig,
)
from unique_swot.services.generation.agentic.prompts.commands.update_existing_section.config import (
    UpdateExistingSectionPromptConfig,
)
from unique_swot.services.generation.context import SWOTComponent
from unique_swot.services.generation.models.base import (
    SWOTReportComponentSection,
)
from unique_swot.services.generation.models.plan import GenerationPlanCommand
from unique_swot.services.generation.models.registry import SWOTReportRegistry
from unique_swot.utils import generate_structured_output

_LOGGER = getLogger(__name__)


async def create_new_section(
    *,
    llm: LMI,
    llm_service: LanguageModelService,
    company_name: str,
    component: SWOTComponent,
    instruction: str,
    command: GenerationPlanCommand,
    fact_id_map: dict[str, str],
    prompts_config: CreateNewSectionPromptConfig,
) -> SWOTReportComponentSection:
    ## Prepare a fact view for the prompt
    facts = {
        key: value
        for key, value in fact_id_map.items()
        if key in command.source_facts_ids
    }

    ## Prepare the prompt
    system_prompt = Template(prompts_config.system_prompt).render(
        component=component,
        company_name=company_name,
    )
    user_message = Template(prompts_config.user_prompt).render(
        facts=facts,
        instruction=instruction,
        model_name=SWOTReportComponentSection.__name__,
        model_schema=json.dumps(
            SWOTReportComponentSection.model_json_schema(), indent=1
        ),
    )

    ## Generate the section
    created_section = await generate_structured_output(
        system_prompt=system_prompt,
        user_message=user_message,
        llm_service=llm_service,
        llm=llm,
        output_model=SWOTReportComponentSection,
    )

    ## Validate the result
    if created_section is None:
        raise FailedToCreateNewSectionException(
            f"Failed to create new section for component {component}"
        )

    _LOGGER.info(f"Created section for component {component}")

    return created_section


async def update_existing_section(
    *,
    llm: LMI,
    llm_service: LanguageModelService,
    company_name: str,
    component: SWOTComponent,
    instruction: str,
    command: GenerationPlanCommand,
    fact_id_map: dict[str, str],
    swot_report_registry: SWOTReportRegistry,
    prompts_config: UpdateExistingSectionPromptConfig,
) -> SWOTReportComponentSection:
    ## Validate the command
    if command.target_section_id is None:
        raise InvalidPlanException(
            f"Target section ID is required for {command.command} command"
        )

    ## Retrieve the existing section
    existing_section = swot_report_registry.retrieve_section(
        id=command.target_section_id
    )

    ## Validate section exists
    if existing_section is None:
        raise SectionNotFoundException(
            f"Section with id {command.target_section_id} not found"
        )

    ## Prepare a fact view for the prompt
    facts = {
        key: value
        for key, value in fact_id_map.items()
        if key in command.source_facts_ids
    }

    ## Prepare the system prompt
    system_prompt = Template(prompts_config.system_prompt).render(
        component=component,
        company_name=company_name,
    )

    ## Prepare the user message
    user_message = Template(prompts_config.user_prompt).render(
        instruction=instruction,
        facts=facts,
        section=existing_section.model_dump_json(indent=1),
        model_name=SWOTReportComponentSection.__name__,
        model_schema=json.dumps(
            SWOTReportComponentSection.model_json_schema(), indent=1
        ),
    )
    ## Generate the section
    updated_section = await generate_structured_output(
        system_prompt=system_prompt,
        user_message=user_message,
        llm_service=llm_service,
        llm=llm,
        output_model=SWOTReportComponentSection,
    )
    ## Validate the result
    if updated_section is None:
        raise FailedToUpdateExistingSectionException(
            f"Failed to update existing section for component {component}"
        )

    _LOGGER.info(f"Updated section for component {component}")

    return updated_section



from logging import getLogger

from unique_toolkit import KnowledgeBaseService
from unique_toolkit.content import Content
from unique_toolkit.content.schemas import (
    ContentInfo,
)

_LOGGER = getLogger(__name__)


async def collect_knowledge_base(
    *,
    knowledge_base_service: KnowledgeBaseService,
    metadata_filter: dict,
) -> list[Content]:
    contents = knowledge_base_service.get_paginated_content_infos(
        metadata_filter=metadata_filter
    )
    contents = await _get_contents_from_paginated_content_infos(
        knowledge_base_service=knowledge_base_service,
        content_infos=contents.content_infos,
    )

    return contents


async def _get_contents_from_paginated_content_infos(
    *,
    knowledge_base_service: KnowledgeBaseService,
    content_infos: list[ContentInfo],
) -> list[Content]:
    contents: list[Content] = []
    for content_info in content_infos:
        content = await _get_chunks_from_content(
            knowledge_base_service=knowledge_base_service,
            content_id=content_info.id,
        )

        if content is None:
            _LOGGER.warning(
                f"No content found for the given content id. Check Ingestion Status: {content_info.id} for more details."
            )
            continue

        if len(content.chunks) == 0:
            _LOGGER.warning(
                f"No chunks found for the given content id. Check Ingestion Status: {content_info.id} for more details."
            )
            continue

        contents.append(content)

    return contents


async def _get_chunks_from_content(
    *,
    knowledge_base_service: KnowledgeBaseService,
    content_id: str,
) -> Content | None:
    contents = await knowledge_base_service.search_contents_async(
        where={"id": {"equals": content_id}}
    )

    if len(contents) == 0:
        _LOGGER.warning(
            f"No content found for the given content id. Check Ingestion Status: {content_id}"
        )
        return

    if len(contents) > 1:
        _LOGGER.warning(
            "Expected exactly one content to be found for the given content id."
            f" Found {len(contents)} contents for the given content id: {content_id}"
            f" Returning the first content."
        )

    return contents[0]



"""Tests for the GenerationAgent."""

from unittest.mock import AsyncMock, Mock, patch

import pytest
from unique_toolkit.content import Content, ContentChunk

from unique_swot.services.generation.agentic.agent import GenerationAgent
from unique_swot.services.generation.context import SWOTComponent
from unique_swot.services.generation.models.base import (
    SWOTReportComponents,
    SWOTReportComponentSection,
    SWOTReportSectionEntry,
)
from unique_swot.services.schemas import SWOTOperation, SWOTPlan, SWOTStepPlan


def _make_content():
    """Helper to create test Content."""
    return Content(
        id="content_1",
        title="Test Document",
        key="test.pdf",
        chunks=[
            ContentChunk(
                id="content_1",
                chunk_id="chunk_1",
                title="Test Document",
                key="test.pdf",
                text="Test content for analysis.",
                start_page=1,
                end_page=1,
                order=0,
            ),
            ContentChunk(
                id="content_1",
                chunk_id="chunk_2",
                title="Test Document",
                key="test.pdf",
                text="More test content.",
                start_page=2,
                end_page=2,
                order=1,
            ),
        ],
    )


def _make_plan(strengths_op=SWOTOperation.GENERATE):
    """Helper to create test SWOTPlan."""
    return SWOTPlan(
        objective="Test analysis",
        strengths=SWOTStepPlan(operation=strengths_op, modify_instruction=None),
        weaknesses=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        opportunities=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        threats=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
    )


@pytest.mark.asyncio
async def test_generation_agent_generate_workflow(
    mock_llm,
    mock_language_model_service,
    mock_swot_report_registry,
    mock_agentic_executor,
    mock_step_notifier,
    mock_progress_notifier,
):
    """Test the complete generate workflow."""
    source_registry = Mock()
    source_registry.register.return_value = "chunk_generated_id"

    config = Mock()
    config.max_tokens_per_extraction_batch = 1000

    agent = GenerationAgent(
        company_name="ACME Corp",
        llm=mock_llm,
        llm_service=mock_language_model_service,
        registry=mock_swot_report_registry,
        executor=mock_agentic_executor,
        config=config,
    )

    content = _make_content()
    plan = _make_plan()

    # Mock the handle_generate_operation function
    with patch(
        "unique_swot.services.generation.agentic.agent.handle_generate_operation",
        new_callable=AsyncMock,
    ) as mock_handle:
        await agent.generate(
            content=content,
            source_registry=source_registry,
            plan=plan,
            step_notifier=mock_step_notifier,
            progress_notifier=mock_progress_notifier,
        )

        # Verify handle_generate_operation was called
        mock_handle.assert_awaited_once()
        call_kwargs = mock_handle.call_args.kwargs
        assert call_kwargs["component"] == SWOTComponent.STRENGTHS
        assert call_kwargs["company_name"] == "ACME Corp"


@pytest.mark.asyncio
async def test_generation_agent_prepares_source_batches(
    mock_llm,
    mock_language_model_service,
    mock_swot_report_registry,
    mock_agentic_executor,
    mock_step_notifier,
    mock_progress_notifier,
):
    """Test that source batches are prepared correctly."""
    source_registry = Mock()
    chunk_ids = ["chunk_id_1", "chunk_id_2"]
    source_registry.register.side_effect = chunk_ids

    config = Mock()
    config.max_tokens_per_extraction_batch = 1000

    agent = GenerationAgent(
        company_name="ACME Corp",
        llm=mock_llm,
        llm_service=mock_language_model_service,
        registry=mock_swot_report_registry,
        executor=mock_agentic_executor,
        config=config,
    )

    content = _make_content()
    plan = _make_plan()

    with patch(
        "unique_swot.services.generation.agentic.agent.handle_generate_operation",
        new_callable=AsyncMock,
    ) as mock_handle:
        await agent.generate(
            content=content,
            source_registry=source_registry,
            plan=plan,
            step_notifier=mock_step_notifier,
            progress_notifier=mock_progress_notifier,
        )

        # Verify source batches were prepared
        call_kwargs = mock_handle.call_args.kwargs
        source_batches = call_kwargs["source_batches"]

        assert len(source_batches) == 2
        assert source_batches[0]["id"] == "chunk_id_1"
        assert source_batches[0]["text"] == "Test content for analysis."
        assert source_batches[1]["id"] == "chunk_id_2"
        assert source_batches[1]["text"] == "More test content."


@pytest.mark.asyncio
async def test_generation_agent_skips_not_requested_operations(
    mock_llm,
    mock_language_model_service,
    mock_swot_report_registry,
    mock_agentic_executor,
    mock_step_notifier,
    mock_progress_notifier,
):
    """Test that NOT_REQUESTED operations are skipped."""
    source_registry = Mock()
    source_registry.register.return_value = "chunk_id"

    config = Mock()
    config.max_tokens_per_extraction_batch = 1000

    agent = GenerationAgent(
        company_name="ACME Corp",
        llm=mock_llm,
        llm_service=mock_language_model_service,
        registry=mock_swot_report_registry,
        executor=mock_agentic_executor,
        config=config,
    )

    content = _make_content()
    # Plan with all components NOT_REQUESTED
    plan = SWOTPlan(
        objective="Test",
        strengths=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        weaknesses=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        opportunities=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        threats=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
    )

    with patch(
        "unique_swot.services.generation.agentic.agent.handle_generate_operation",
        new_callable=AsyncMock,
    ) as mock_handle:
        await agent.generate(
            content=content,
            source_registry=source_registry,
            plan=plan,
            step_notifier=mock_step_notifier,
            progress_notifier=mock_progress_notifier,
        )

        # Should not be called for any component
        mock_handle.assert_not_awaited()


@pytest.mark.asyncio
async def test_generation_agent_handles_modify_operation(
    mock_llm,
    mock_language_model_service,
    mock_swot_report_registry,
    mock_agentic_executor,
    mock_step_notifier,
    mock_progress_notifier,
):
    """Test that MODIFY operations are handled (currently uses GENERATE)."""
    source_registry = Mock()
    source_registry.register.return_value = "chunk_id"

    config = Mock()
    config.max_tokens_per_extraction_batch = 1000

    agent = GenerationAgent(
        company_name="ACME Corp",
        llm=mock_llm,
        llm_service=mock_language_model_service,
        registry=mock_swot_report_registry,
        executor=mock_agentic_executor,
        config=config,
    )

    content = _make_content()
    plan = SWOTPlan(
        objective="Update analysis",
        strengths=SWOTStepPlan(
            operation=SWOTOperation.MODIFY, modify_instruction="Update with new data"
        ),
        weaknesses=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        opportunities=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        threats=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
    )

    with patch(
        "unique_swot.services.generation.agentic.agent.handle_generate_operation",
        new_callable=AsyncMock,
    ) as mock_handle:
        await agent.generate(
            content=content,
            source_registry=source_registry,
            plan=plan,
            step_notifier=mock_step_notifier,
            progress_notifier=mock_progress_notifier,
        )

        # Should be called once for MODIFY (currently treated as GENERATE)
        mock_handle.assert_awaited_once()


@pytest.mark.asyncio
async def test_generation_agent_sends_notifications(
    mock_llm,
    mock_language_model_service,
    mock_swot_report_registry,
    mock_agentic_executor,
    mock_step_notifier,
    mock_progress_notifier,
):
    """Test that notifications are sent during generation."""
    source_registry = Mock()
    source_registry.register.return_value = "chunk_id"

    config = Mock()
    config.max_tokens_per_extraction_batch = 1000

    agent = GenerationAgent(
        company_name="ACME Corp",
        llm=mock_llm,
        llm_service=mock_language_model_service,
        registry=mock_swot_report_registry,
        executor=mock_agentic_executor,
        config=config,
    )

    content = _make_content()
    plan = _make_plan()

    with patch(
        "unique_swot.services.generation.agentic.agent.handle_generate_operation",
        new_callable=AsyncMock,
    ):
        await agent.generate(
            content=content,
            source_registry=source_registry,
            plan=plan,
            step_notifier=mock_step_notifier,
            progress_notifier=mock_progress_notifier,
        )

        # Verify notifications were sent
        assert mock_step_notifier.notify.await_count >= 2  # Start and end notifications


def test_generation_agent_get_reports(
    mock_llm, mock_language_model_service, mock_agentic_executor
):
    """Test get_reports returns correct structure."""
    registry = Mock()
    registry.retrieve_component_sections.return_value = [
        SWOTReportComponentSection(
            h2="Test", entries=[SWOTReportSectionEntry(preview="P", content="C")]
        )
    ]

    config = Mock()
    config.max_tokens_per_extraction_batch = 1000

    agent = GenerationAgent(
        company_name="ACME Corp",
        llm=mock_llm,
        llm_service=mock_language_model_service,
        registry=registry,
        executor=mock_agentic_executor,
        config=config,
    )

    reports = agent.get_reports()

    assert isinstance(reports, SWOTReportComponents)
    assert len(reports.strengths) == 1
    assert len(reports.weaknesses) == 1
    assert len(reports.opportunities) == 1
    assert len(reports.threats) == 1


def test_generation_agent_notification_title_property(
    mock_llm,
    mock_language_model_service,
    mock_swot_report_registry,
    mock_agentic_executor,
):
    """Test notification_title property getter and setter."""
    config = Mock()
    config.max_tokens_per_extraction_batch = 1000

    agent = GenerationAgent(
        company_name="ACME Corp",
        llm=mock_llm,
        llm_service=mock_language_model_service,
        registry=mock_swot_report_registry,
        executor=mock_agentic_executor,
        config=config,
    )

    # Default title
    assert agent.notification_title == "Generating SWOT report"

    # Set new title
    agent.notification_title = "Custom Title"
    assert agent.notification_title == "Custom Title"



"""Shared test fixtures and configurations for SWOT analysis tests."""

from unittest.mock import AsyncMock, Mock

import pytest
from unique_toolkit.content import Content, ContentChunk

from unique_swot.services.generation.models.base import (
    SWOTReportComponents,
    SWOTReportComponentSection,
    SWOTReportSectionEntry,
)
from unique_swot.services.schemas import SWOTOperation, SWOTPlan, SWOTStepPlan

# ============================================================================
# External Service Mocks
# ============================================================================


@pytest.fixture
def mock_knowledge_base_service():
    """Mock KnowledgeBaseService for testing."""
    service = Mock()
    service.search_content.return_value = []
    service.upload_content_from_bytes.return_value = Mock(id="test_content_id")
    service.download_content_to_bytes.return_value = b'{"test": "data"}'
    return service


@pytest.fixture
def mock_language_model_service():
    """Mock LanguageModelService for testing."""
    service = Mock()
    service.generate_text = AsyncMock(return_value="Generated text")
    service.complete_async = AsyncMock()
    return service


@pytest.fixture
def mock_short_term_memory_service():
    """Mock ShortTermMemoryService for testing."""
    service = Mock()
    service.create_memory.return_value = None
    service.find_latest_memory.return_value = None
    return service


@pytest.fixture
def mock_chat_service():
    """Mock ChatService for testing."""
    service = Mock()
    service.assistant_message_id = "test_message_id"
    service.create_message_log.return_value = Mock(message_log_id="test_log_id")
    service.create_message_execution.return_value = None
    service.update_message_execution.return_value = None
    service.modify_assistant_message.return_value = None
    service.upload_to_chat_from_bytes.return_value = Mock(
        id="uploaded_content_id", title="Test Document", key="test.docx"
    )
    return service


@pytest.fixture
def mock_quartr_service():
    """Mock QuartrService for testing."""
    service = Mock()
    service.get_earnings_calls = AsyncMock(return_value=[])
    return service


@pytest.fixture
def mock_docx_generator():
    """Mock DocxGeneratorService for testing."""
    service = Mock()
    service.generate.return_value = b"fake docx content"
    return service


# ============================================================================
# Content & Chunk Fixtures
# ============================================================================


@pytest.fixture
def sample_content_chunk():
    """Create a sample ContentChunk for testing."""
    return ContentChunk(
        id="content_123",
        chunk_id="chunk_456",
        title="Test Document",
        key="test_doc.pdf",
        text="This is sample content for testing.",
        start_page=1,
        end_page=2,
        order=0,
    )


@pytest.fixture
def sample_content(sample_content_chunk):
    """Create a sample Content object with chunks."""
    chunk2 = ContentChunk(
        id="content_123",
        chunk_id="chunk_789",
        title="Test Document",
        key="test_doc.pdf",
        text="This is more sample content.",
        start_page=3,
        end_page=4,
        order=1,
    )
    return Content(
        id="content_123",
        title="Test Document",
        key="test_doc.pdf",
        chunks=[sample_content_chunk, chunk2],
    )


@pytest.fixture
def sample_contents(sample_content):
    """Create a list of sample Content objects."""
    content2 = Content(
        id="content_456",
        title="Second Document",
        key="second_doc.pdf",
        chunks=[
            ContentChunk(
                id="content_456",
                chunk_id="chunk_abc",
                title="Second Document",
                key="second_doc.pdf",
                text="Second document content.",
                start_page=1,
                end_page=1,
                order=0,
            )
        ],
    )
    return [sample_content, content2]


# ============================================================================
# SWOT Plan Fixtures
# ============================================================================


@pytest.fixture
def sample_swot_step_plan():
    """Create a sample SWOTStepPlan for testing."""
    return SWOTStepPlan(
        operation=SWOTOperation.GENERATE,
        modify_instruction=None,
    )


@pytest.fixture
def sample_swot_plan(sample_swot_step_plan):
    """Create a sample SWOTPlan for testing."""
    return SWOTPlan(
        objective="Analyze company X's competitive position",
        strengths=sample_swot_step_plan,
        weaknesses=sample_swot_step_plan,
        opportunities=sample_swot_step_plan,
        threats=sample_swot_step_plan,
    )


@pytest.fixture
def sample_modify_swot_plan():
    """Create a sample SWOTPlan with modify operations."""
    return SWOTPlan(
        objective="Update existing SWOT analysis",
        strengths=SWOTStepPlan(
            operation=SWOTOperation.MODIFY,
            modify_instruction="Update with new data",
        ),
        weaknesses=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED,
            modify_instruction=None,
        ),
        opportunities=SWOTStepPlan(
            operation=SWOTOperation.GENERATE,
            modify_instruction=None,
        ),
        threats=SWOTStepPlan(
            operation=SWOTOperation.MODIFY,
            modify_instruction="Add recent threats",
        ),
    )


@pytest.fixture
def sample_plan_strengths_only():
    """Create a SWOTPlan with only strengths requested."""
    return SWOTPlan(
        objective="Analyze company strengths",
        strengths=SWOTStepPlan(
            operation=SWOTOperation.GENERATE, modify_instruction=None
        ),
        weaknesses=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        opportunities=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        threats=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
    )


# ============================================================================
# SWOT Report Fixtures
# ============================================================================


@pytest.fixture
def sample_report_section():
    """Create a sample SWOTReportComponentSection."""
    return SWOTReportComponentSection(
        h2="Strong Market Position",
        entries=[
            SWOTReportSectionEntry(
                preview="Leading market share",
                content="The company holds a 35% market share in the technology sector [chunk_a].",
            ),
            SWOTReportSectionEntry(
                preview="Brand recognition",
                content="Strong brand recognition globally [chunk_b].",
            ),
        ],
    )


@pytest.fixture
def sample_report_components(sample_report_section):
    """Create a sample SWOTReportComponents."""
    return SWOTReportComponents(
        strengths=[sample_report_section],
        weaknesses=[],
        opportunities=[],
        threats=[],
    )


# ============================================================================
# Service Component Mocks
# ============================================================================


@pytest.fixture
def mock_step_notifier():
    """Mock StepNotifier for testing."""
    notifier = Mock()
    notifier.notify = AsyncMock()
    notifier.get_total_number_of_references.return_value = 0
    return notifier


@pytest.fixture
def mock_swot_memory_service():
    """Mock SwotMemoryService for testing."""
    service = Mock()
    service.get.return_value = None
    service.set.return_value = None
    return service


@pytest.fixture
def mock_content_chunk_registry():
    """Mock ContentChunkRegistry for testing."""
    registry = Mock()
    registry.register.return_value = "chunk_generated_id"
    registry.retrieve.return_value = None
    registry.save.return_value = None
    return registry


@pytest.fixture
def mock_swot_report_registry():
    """Mock SWOTReportRegistry for testing."""
    registry = Mock()
    registry.register_section.return_value = "section_id_123"
    registry.retrieve_section.return_value = None
    registry.retrieve_component_sections.return_value = []
    registry.retrieve_sections_for_component.return_value = "{}"
    return registry


@pytest.fixture
def mock_agentic_executor():
    """Mock AgenticPlanExecutor for testing."""
    executor = Mock()
    executor.execute = AsyncMock(return_value=None)
    return executor


@pytest.fixture
def mock_citation_manager():
    """Mock CitationManager for testing."""
    manager = Mock()
    manager.add_citations_to_report.return_value = "Report with citations"
    manager.get_references.return_value = []
    manager.get_citations_for_docx.return_value = []
    manager.get_referenced_content_chunks.return_value = []
    return manager


@pytest.fixture
def mock_llm():
    """Mock LLM (Language Model Interface) for testing."""
    llm = Mock()
    llm.name = "test-model"
    # encoder_name should be an object with a .value attribute (enum-like)
    encoder_name_mock = Mock()
    encoder_name_mock.value = "cl100k_base"
    llm.encoder_name = encoder_name_mock
    return llm


@pytest.fixture
def mock_progress_notifier():
    """Mock ProgressNotifier for testing."""
    notifier = Mock()
    notifier.update = AsyncMock()
    notifier.increment = AsyncMock()
    notifier.step_size = 0
    return notifier


# ============================================================================
# Protocol Implementation Mocks
# ============================================================================


@pytest.fixture
def mock_source_collector():
    """Mock SourceCollector protocol for testing."""
    collector = Mock()
    collector.collect = AsyncMock(return_value=[])
    return collector


@pytest.fixture
def mock_source_selector():
    """Mock SourceSelector protocol for testing."""
    selector = Mock()
    selector.select = AsyncMock()
    return selector


@pytest.fixture
def mock_source_iterator():
    """Mock SourceIterator protocol for testing."""
    iterator = Mock()

    async def _iterate(contents, step_notifier):
        for content in contents:
            yield content

    iterator.iterate = _iterate
    return iterator


@pytest.fixture
def mock_reporting_agent():
    """Mock ReportingAgent protocol for testing."""
    agent = Mock()
    agent.generate = AsyncMock()
    agent.get_reports.return_value = SWOTReportComponents(
        strengths=[], weaknesses=[], opportunities=[], threats=[]
    )
    return agent



"""Tests for the SourceIterationAgent."""

from unittest.mock import AsyncMock, Mock, patch

import pytest
from unique_toolkit.content import Content, ContentChunk

from unique_swot.services.source_management.iteration.agent import SourceIterationAgent
from unique_swot.services.source_management.iteration.schema import (
    SourceIterationResult,
    SourceIterationResults,
)


def _make_content(content_id="content_1", title="Test Doc"):
    """Helper to create test Content."""
    return Content(
        id=content_id,
        title=title,
        key=f"{title}.pdf",
        chunks=[
            ContentChunk(
                id=content_id,
                chunk_id="chunk_1",
                title=title,
                key=f"{title}.pdf",
                text=f"Content for {title}",
                start_page=1,
                end_page=1,
                order=0,
            )
        ],
    )


@pytest.mark.asyncio
async def test_iteration_agent_returns_ordered_sources(
    mock_language_model_service, mock_llm, mock_step_notifier
):
    """Test that agent returns sources in LLM-specified order."""
    config = Mock()
    config.prompt_config.system_prompt = "System prompt with {{objective}}"
    # Use a template that includes the content IDs
    config.prompt_config.user_prompt = """
{% for content in contents %}
Source ID: {{ content.id }}
Title: {{ content.document_title }}
Chunks: {{ content.chunks }}
{% endfor %}
"""
    config.prompt_config.objective = "Sort by relevance"
    config.max_number_of_selected_chunks = 5

    agent = SourceIterationAgent(
        llm_service=mock_language_model_service,
        llm=mock_llm,
        source_iteration_config=config,
    )

    contents = [
        _make_content("c1", "Doc A"),
        _make_content("c2", "Doc B"),
        _make_content("c3", "Doc C"),
    ]

    # Mock LLM to return reversed order
    # Note: The IDs will be generated, so we need to capture them
    def capture_user_prompt(*args, **kwargs):
        # Extract IDs from the user prompt
        user_prompt = kwargs.get("user_message", "")
        import re

        ids = re.findall(r"content_[a-zA-Z0-9]+", user_prompt)

        # Ensure we have at least 3 IDs
        if len(ids) < 3:
            return SourceIterationResults(
                ordered_sources=[],
                results_summary="Not enough sources",
            )

        # Return ordered results (reversed)
        return SourceIterationResults(
            ordered_sources=[
                SourceIterationResult(id=ids[2], order=1),
                SourceIterationResult(id=ids[1], order=2),
                SourceIterationResult(id=ids[0], order=3),
            ],
            results_summary="Sorted by relevance",
        )

    with patch(
        "unique_swot.services.source_management.iteration.agent.generate_structured_output",
        new_callable=AsyncMock,
        side_effect=capture_user_prompt,
    ):
        iterator = await agent.iterate(
            contents=contents,
            step_notifier=mock_step_notifier,
        )

        # Collect results from async iterator
        results = []
        async for content in iterator:
            results.append(content)

        # Should return all 3 contents
        assert len(results) == 3
        # Order should be reversed (C, B, A)
        assert results[0].title == "Doc C"
        assert results[1].title == "Doc B"
        assert results[2].title == "Doc A"


@pytest.mark.asyncio
async def test_iteration_agent_handles_llm_failure(
    mock_language_model_service, mock_llm, mock_step_notifier
):
    """Test that agent falls back to original order on LLM failure."""
    config = Mock()
    config.prompt_config.system_prompt = "System prompt with {{objective}}"
    config.prompt_config.user_prompt = """
{% for content in contents %}
Source ID: {{ content.id }}
{% endfor %}
"""
    config.prompt_config.objective = "Sort by relevance"
    config.max_number_of_selected_chunks = 5

    agent = SourceIterationAgent(
        llm_service=mock_language_model_service,
        llm=mock_llm,
        source_iteration_config=config,
    )

    contents = [
        _make_content("c1", "Doc A"),
        _make_content("c2", "Doc B"),
    ]

    # Mock LLM failure (returns None)
    with patch(
        "unique_swot.services.source_management.iteration.agent.generate_structured_output",
        new_callable=AsyncMock,
        return_value=None,
    ):
        iterator = await agent.iterate(
            contents=contents,
            step_notifier=mock_step_notifier,
        )

        results = []
        async for content in iterator:
            results.append(content)

        # Should return all contents in original order
        assert len(results) == 2
        assert results[0].title == "Doc A"
        assert results[1].title == "Doc B"


@pytest.mark.asyncio
async def test_iteration_agent_handles_missed_documents(
    mock_language_model_service, mock_llm, mock_step_notifier
):
    """Test that agent includes documents missed by LLM."""
    config = Mock()
    config.prompt_config.system_prompt = "System prompt with {{objective}}"
    config.prompt_config.user_prompt = """
{% for content in contents %}
Source ID: {{ content.id }}
{% endfor %}
"""
    config.prompt_config.objective = "Sort by relevance"
    config.max_number_of_selected_chunks = 5

    agent = SourceIterationAgent(
        llm_service=mock_language_model_service,
        llm=mock_llm,
        source_iteration_config=config,
    )

    contents = [
        _make_content("c1", "Doc A"),
        _make_content("c2", "Doc B"),
        _make_content("c3", "Doc C"),
    ]

    # Mock LLM to return only 2 documents (missing one)
    def return_partial_results(*args, **kwargs):
        user_prompt = kwargs.get("user_message", "")
        import re

        ids = re.findall(r"content_[a-zA-Z0-9]+", user_prompt)

        # Ensure we have at least 2 IDs
        if len(ids) < 2:
            return SourceIterationResults(
                ordered_sources=[],
                results_summary="Not enough sources",
            )

        # Only return first 2 IDs
        return SourceIterationResults(
            ordered_sources=[
                SourceIterationResult(id=ids[0], order=1),
                SourceIterationResult(id=ids[1], order=2),
            ],
            results_summary="Sorted",
        )

    with patch(
        "unique_swot.services.source_management.iteration.agent.generate_structured_output",
        new_callable=AsyncMock,
        side_effect=return_partial_results,
    ):
        iterator = await agent.iterate(
            contents=contents,
            step_notifier=mock_step_notifier,
        )

        results = []
        async for content in iterator:
            results.append(content)

        # Should return all 3 contents (missed one first, then ordered ones)
        assert len(results) == 3


@pytest.mark.asyncio
async def test_iteration_agent_sends_notifications(
    mock_language_model_service, mock_llm, mock_step_notifier
):
    """Test that notifications are sent during iteration."""
    config = Mock()
    config.prompt_config.system_prompt = "System prompt with {{objective}}"
    config.prompt_config.user_prompt = """
{% for content in contents %}
Source ID: {{ content.id }}
{% endfor %}
"""
    config.prompt_config.objective = "Sort by relevance"
    config.max_number_of_selected_chunks = 5

    agent = SourceIterationAgent(
        llm_service=mock_language_model_service,
        llm=mock_llm,
        source_iteration_config=config,
    )

    contents = [_make_content("c1", "Doc A")]

    mock_result = SourceIterationResults(
        ordered_sources=[],
        results_summary="Sorted",
    )

    with patch(
        "unique_swot.services.source_management.iteration.agent.generate_structured_output",
        new_callable=AsyncMock,
        return_value=mock_result,
    ):
        await agent.iterate(
            contents=contents,
            step_notifier=mock_step_notifier,
        )

        # Verify notifications were sent
        assert mock_step_notifier.notify.await_count >= 2  # Start and end


@pytest.mark.asyncio
async def test_iteration_agent_limits_chunks(
    mock_language_model_service, mock_llm, mock_step_notifier
):
    """Test that agent respects max_number_of_selected_chunks."""
    config = Mock()
    config.prompt_config.system_prompt = "System prompt with {{objective}}"
    # Use a template that includes the chunks
    config.prompt_config.user_prompt = """
{% for content in contents %}
Source ID: {{ content.id }}
Title: {{ content.document_title }}
Content: {{ content.chunks }}
{% endfor %}
"""
    config.prompt_config.objective = "Sort by relevance"
    config.max_number_of_selected_chunks = 2

    agent = SourceIterationAgent(
        llm_service=mock_language_model_service,
        llm=mock_llm,
        source_iteration_config=config,
    )

    # Create content with many chunks
    content = Content(
        id="content_1",
        title="Test Document",
        key="test.pdf",
        chunks=[
            ContentChunk(
                id="content_1",
                chunk_id=f"chunk_{i}",
                title="Test Document",
                key="test.pdf",
                text=f"Chunk {i} content",
                start_page=i,
                end_page=i,
                order=i,
            )
            for i in range(10)
        ],
    )

    mock_result = SourceIterationResults(
        ordered_sources=[],
        results_summary="Sorted",
    )

    with patch(
        "unique_swot.services.source_management.iteration.agent.generate_structured_output",
        new_callable=AsyncMock,
        return_value=mock_result,
    ) as mock_generate:
        await agent.iterate(
            contents=[content],
            step_notifier=mock_step_notifier,
        )

        # Verify the call was made
        mock_generate.assert_awaited_once()

        # Check that the user prompt only includes limited chunks
        call_kwargs = mock_generate.call_args.kwargs
        user_prompt = call_kwargs["user_message"]
        # Should only have 2 chunks worth of content (Chunk 0 and Chunk 1)
        assert user_prompt.count("Chunk") == 2


@pytest.mark.asyncio
async def test_iteration_agent_empty_contents(
    mock_language_model_service, mock_llm, mock_step_notifier
):
    """Test that agent handles empty contents list."""
    config = Mock()
    config.prompt_config.system_prompt = "System prompt with {{objective}}"
    config.prompt_config.user_prompt = "User prompt"
    config.prompt_config.objective = "Sort by relevance"
    config.max_number_of_selected_chunks = 5

    agent = SourceIterationAgent(
        llm_service=mock_language_model_service,
        llm=mock_llm,
        source_iteration_config=config,
    )

    mock_result = SourceIterationResults(
        ordered_sources=[],
        results_summary="No sources to sort",
    )

    with patch(
        "unique_swot.services.source_management.iteration.agent.generate_structured_output",
        new_callable=AsyncMock,
        return_value=mock_result,
    ):
        iterator = await agent.iterate(
            contents=[],
            step_notifier=mock_step_notifier,
        )

        results = []
        async for content in iterator:
            results.append(content)

        # Should return empty list
        assert len(results) == 0



import json
from logging import getLogger

from unique_toolkit import LanguageModelService
from unique_toolkit._common.validators import LMI
from unique_toolkit.content import Content

from unique_swot.services.generation.agentic.config import AgenticGeneratorConfig
from unique_swot.services.generation.agentic.executor import AgenticPlanExecutor
from unique_swot.services.generation.agentic.operations import (
    handle_generate_operation,
)
from unique_swot.services.generation.agentic.utils import (
    batch_sequence_generator,
    create_batch_notification,
)
from unique_swot.services.generation.context import SWOTComponent
from unique_swot.services.generation.models.base import (
    SWOTReportComponents,
)
from unique_swot.services.generation.models.registry import SWOTReportRegistry
from unique_swot.services.orchestrator.service import (
    ProgressNotifier,
    SourceRegistry,
    StepNotifier,
)
from unique_swot.services.schemas import SWOTOperation, SWOTPlan
from unique_swot.utils import (
    convert_content_chunk_to_reference,
    get_content_chunk_title,
)

_LOGGER = getLogger(__name__)


class GenerationAgent:
    def __init__(
        self,
        *,
        company_name: str,
        llm: LMI,
        llm_service: LanguageModelService,
        registry: SWOTReportRegistry,
        executor: AgenticPlanExecutor,
        config: AgenticGeneratorConfig,
    ):
        self._llm_service = llm_service
        self._registry = registry
        self._company_name = company_name
        self._swot_report_registry = registry
        self._llm = llm
        self._executor = executor
        self._title = "Generating SWOT report"
        self._config = config

    @property
    def notification_title(self) -> str:
        return self._title

    @notification_title.setter
    def notification_title(self, value: str):
        self._title = value

    async def generate(
        self,
        *,
        content: Content,
        source_registry: SourceRegistry,
        plan: SWOTPlan,
        step_notifier: StepNotifier,
        progress_notifier: ProgressNotifier,
    ):
        _LOGGER.info("Starting SWOT Analysis generation Agent")
        # Prepare the source splits
        components_step = plan.convert_plan_to_list_of_steps()
        source_batches = self._prepare_source_batches(
            content=content, source_registry=source_registry
        )

        document_title = get_content_chunk_title(content)
        self.notification_title = f"**Processing `{document_title}`**"

        document_reference = convert_content_chunk_to_reference(
            content_or_chunk=content,
        )

        await step_notifier.notify(
            title=self.notification_title,
            sources=[document_reference],
            progress=0,
        )

        fraction_step = 1 / len(components_step)

        batches_generator = batch_sequence_generator(
            encoder_name=self._llm.encoder_name.value,
            source_batches=source_batches,
            max_tokens_per_extraction_batch=self._config.max_tokens_per_extraction_batch,
            serializer=json.dumps,
        )

        batches = list(batches_generator)
        total_batches = len(batches)

        if total_batches > 1:
            await step_notifier.notify(
                title=self.notification_title,
                description=f"This document is too large to extract. It will be split into {len(batches)} batches.",
            )

        for index, (component, step) in enumerate(components_step):
            for i, batch in enumerate(batches, start=1):
                await step_notifier.notify(
                    title=self.notification_title,
                    progress=int(index / len(components_step) * 100),
                    description=create_batch_notification(
                        component=component.value,
                        batch_index=i,
                        total_batches=len(batches),
                    ),
                )
                match step.operation:
                    case SWOTOperation.GENERATE:
                        await handle_generate_operation(
                            component=component,
                            source_batches=batch,
                            step_notifier=step_notifier,
                            company_name=self._company_name,
                            llm=self._llm,
                            llm_service=self._llm_service,
                            notification_title=self.notification_title,
                            swot_report_registry=self._swot_report_registry,
                            executor=self._executor,
                            config=self._config,
                        )
                    case SWOTOperation.MODIFY:
                        _LOGGER.warning(
                            "Modification operation not supported yet. Using generate operation instead."
                        )
                        await handle_generate_operation(
                            component=component,
                            source_batches=batch,
                            step_notifier=step_notifier,
                            company_name=self._company_name,
                            llm=self._llm,
                            llm_service=self._llm_service,
                            notification_title=self.notification_title,
                            swot_report_registry=self._swot_report_registry,
                            executor=self._executor,
                            config=self._config,
                        )
                    case SWOTOperation.NOT_REQUESTED:
                        continue
                    case _:
                        raise ValueError(f"Invalid operation: {step.operation}")

            await progress_notifier.increment(fraction=fraction_step)

        await step_notifier.notify(
            title=self.notification_title,
            progress=100,
            description=f"Completed processing {document_title}!",
            completed=True,
        )

    def _prepare_source_batches(
        self, *, content: Content, source_registry: SourceRegistry
    ) -> list[dict[str, str]]:
        """
        Prepare source chunks for prompting while keeping chunk ids for citation.

        Each entry contains the chunk id and text so downstream prompts can
        instruct the model to cite using [chunk_id].
        """
        ## Register the chunks
        chunk_ids: list[str] = []
        ordered_chunks = sorted(content.chunks, key=lambda x: x.order)
        for chunk in ordered_chunks:
            chunk_id = source_registry.register(chunk=chunk)
            chunk_ids.append(chunk_id)

        return [
            {"id": chunk_id, "text": chunk.text}
            for chunk_id, chunk in zip(chunk_ids, ordered_chunks)
        ]

    def get_reports(self) -> SWOTReportComponents:
        return SWOTReportComponents(
            strengths=self._swot_report_registry.retrieve_component_sections(
                component=SWOTComponent.STRENGTHS
            ),
            weaknesses=self._swot_report_registry.retrieve_component_sections(
                component=SWOTComponent.WEAKNESSES
            ),
            opportunities=self._swot_report_registry.retrieve_component_sections(
                component=SWOTComponent.OPPORTUNITIES
            ),
            threats=self._swot_report_registry.retrieve_component_sections(
                component=SWOTComponent.THREATS
            ),
        )



"""Tests for the AgenticPlanExecutor."""

import asyncio

import pytest

from unique_swot.services.generation.agentic.executor import (
    AgenticPlanExecutor,
    ExecutionMode,
)


async def _success_task(value: int) -> int:
    """A simple async task that returns its input."""
    await asyncio.sleep(0.01)
    return value * 2


async def _failing_task() -> None:
    """A task that always raises an exception."""
    await asyncio.sleep(0.01)
    raise ValueError("Task failed")


@pytest.mark.asyncio
async def test_executor_sequential_mode_success():
    """Test sequential execution with successful tasks."""
    executor = AgenticPlanExecutor(execution_mode=ExecutionMode.SEQUENTIAL)

    executor.add(_success_task, 1)
    executor.add(_success_task, 2)
    executor.add(_success_task, 3)

    results = await executor.run()

    assert len(results) == 3
    assert results[0] == 2
    assert results[1] == 4
    assert results[2] == 6


@pytest.mark.asyncio
async def test_executor_sequential_mode_with_failure():
    """Test sequential execution with a failing task."""
    executor = AgenticPlanExecutor(execution_mode=ExecutionMode.SEQUENTIAL)

    executor.add(_success_task, 1)
    executor.add(_failing_task)
    executor.add(_success_task, 3)

    results = await executor.run()

    assert len(results) == 3
    assert results[0] == 2
    assert isinstance(results[1], ValueError)
    assert results[2] == 6


@pytest.mark.asyncio
async def test_executor_concurrent_mode_success():
    """Test concurrent execution with successful tasks."""
    executor = AgenticPlanExecutor(
        execution_mode=ExecutionMode.CONCURRENT, max_concurrent_tasks=2
    )

    executor.add(_success_task, 1)
    executor.add(_success_task, 2)
    executor.add(_success_task, 3)

    results = await executor.run()

    assert len(results) == 3
    # Results should all be successful (order may vary in concurrent mode)
    assert set(results) == {2, 4, 6}


@pytest.mark.asyncio
async def test_executor_concurrent_mode_with_failure():
    """Test concurrent execution with a failing task."""
    executor = AgenticPlanExecutor(
        execution_mode=ExecutionMode.CONCURRENT, max_concurrent_tasks=2
    )

    executor.add(_success_task, 1)
    executor.add(_failing_task)
    executor.add(_success_task, 3)

    results = await executor.run()

    assert len(results) == 3
    # Two successful results and one exception
    successful_results = [r for r in results if not isinstance(r, Exception)]
    failed_results = [r for r in results if isinstance(r, Exception)]

    assert len(successful_results) == 2
    assert len(failed_results) == 1
    assert isinstance(failed_results[0], ValueError)


@pytest.mark.asyncio
async def test_executor_respects_max_concurrent_tasks():
    """Test that max_concurrent_tasks limit is respected."""
    call_times = []

    async def _timed_task(task_id: int):
        call_times.append((task_id, "start"))
        await asyncio.sleep(0.05)
        call_times.append((task_id, "end"))
        return task_id

    executor = AgenticPlanExecutor(
        execution_mode=ExecutionMode.CONCURRENT, max_concurrent_tasks=2
    )

    for i in range(4):
        executor.add(_timed_task, i)

    results = await executor.run()

    # All tasks should complete
    assert len(results) == 4
    assert set(results) == {0, 1, 2, 3}


@pytest.mark.asyncio
async def test_executor_clears_queue_after_run():
    """Test that the queue is cleared after execution."""
    executor = AgenticPlanExecutor(execution_mode=ExecutionMode.SEQUENTIAL)

    executor.add(_success_task, 1)
    executor.add(_success_task, 2)

    results1 = await executor.run()
    assert len(results1) == 2

    # Queue should be empty, running again should return empty results
    results2 = await executor.run()
    assert len(results2) == 0


@pytest.mark.asyncio
async def test_executor_with_kwargs():
    """Test executor with keyword arguments."""

    async def _task_with_kwargs(a: int, b: int = 0) -> int:
        return a + b

    executor = AgenticPlanExecutor(execution_mode=ExecutionMode.SEQUENTIAL)

    executor.add(_task_with_kwargs, 1, b=2)
    executor.add(_task_with_kwargs, 3, b=4)

    results = await executor.run()

    assert results[0] == 3
    assert results[1] == 7


@pytest.mark.asyncio
async def test_executor_empty_queue():
    """Test executor with no tasks added."""
    executor = AgenticPlanExecutor(execution_mode=ExecutionMode.SEQUENTIAL)

    results = await executor.run()

    assert len(results) == 0


@pytest.mark.asyncio
async def test_executor_multiple_runs():
    """Test that executor can be reused for multiple runs."""
    executor = AgenticPlanExecutor(execution_mode=ExecutionMode.SEQUENTIAL)

    # First run
    executor.add(_success_task, 1)
    results1 = await executor.run()
    assert len(results1) == 1
    assert results1[0] == 2

    # Second run with different tasks
    executor.add(_success_task, 5)
    executor.add(_success_task, 10)
    results2 = await executor.run()
    assert len(results2) == 2
    assert results2[0] == 10
    assert results2[1] == 20


@pytest.mark.asyncio
async def test_executor_default_max_concurrent_tasks():
    """Test that default max_concurrent_tasks is set correctly."""
    executor = AgenticPlanExecutor(execution_mode=ExecutionMode.CONCURRENT)

    # Default should be 10
    assert executor._max_concurrent_tasks == 10

    # Add more than 10 tasks
    for i in range(15):
        executor.add(_success_task, i)

    results = await executor.run()

    # All tasks should complete
    assert len(results) == 15



from enum import StrEnum

from pydantic import Field

from unique_swot.utils import StructuredOutputResult, StructuredOutputWithNotification


class GenerationPlanCommandType(StrEnum):
    CREATE_SECTION = "create_section"
    UPDATE_SECTION = "update_section"


class GenerationPlanCommand(StructuredOutputResult):
    reasoning: str = Field(
        description="The reasoning that led to the command. Need to be concise and backed with evidence."
    )
    command: GenerationPlanCommandType

    instruction: str = Field(
        description="The complete instruction to be executed. Should contain all the necessary details to ensure the command is executed correctly."
    )
    target_section_id: str | None = Field(
        description="The ID of the section to be executed. This is only used if the command is update_section"
    )
    source_facts_ids: list[str] = Field(
        description="The IDs of the source facts of similar nature that need to go into the section to be updated or created."
    )


class GenerationPlan(StructuredOutputWithNotification):
    commands: list[GenerationPlanCommand] = Field(
        description="The commands to be executed. The commands should be executed in order. Each command should target a specific section to be updated or created."
    )



"""Tests for the SWOT Orchestrator service."""

from unittest.mock import AsyncMock, Mock

import pytest
from unique_toolkit.content import Content, ContentChunk

from unique_swot.services.generation.models.base import (
    SWOTReportComponents,
    SWOTReportComponentSection,
    SWOTReportSectionEntry,
)
from unique_swot.services.orchestrator.service import SWOTOrchestrator
from unique_swot.services.schemas import SWOTOperation, SWOTPlan, SWOTStepPlan
from unique_swot.services.source_management.selection.schema import (
    SourceSelectionResult,
)


def _make_content(content_id="content_1", title="Test Doc"):
    """Helper to create test Content."""
    return Content(
        id=content_id,
        title=title,
        key=f"{title}.pdf",
        chunks=[
            ContentChunk(
                id=content_id,
                chunk_id="chunk_1",
                title=title,
                key=f"{title}.pdf",
                text="Test content",
                start_page=1,
                end_page=1,
                order=0,
            )
        ],
    )


def _make_plan(strengths_op=SWOTOperation.GENERATE):
    """Helper to create test SWOTPlan."""
    return SWOTPlan(
        objective="Test objective",
        strengths=SWOTStepPlan(operation=strengths_op, modify_instruction=None),
        weaknesses=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        opportunities=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        threats=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
    )


def _make_report_components():
    """Helper to create test SWOTReportComponents."""
    return SWOTReportComponents(
        strengths=[
            SWOTReportComponentSection(
                h2="Test Strength",
                entries=[
                    SWOTReportSectionEntry(
                        preview="Preview", content="Content [chunk_1]"
                    )
                ],
            )
        ],
        weaknesses=[],
        opportunities=[],
        threats=[],
    )


@pytest.mark.asyncio
async def test_orchestrator_full_workflow():
    """Test the complete orchestration workflow: collect → iterate → select → generate."""
    # Setup mocks
    step_notifier = Mock()
    step_notifier.notify = AsyncMock()

    contents = [_make_content("content_1", "Doc1"), _make_content("content_2", "Doc2")]

    source_collector = Mock()
    source_collector.collect = AsyncMock(return_value=contents)

    source_iterator = Mock()

    async def _iterate_impl(contents, step_notifier):
        for content in contents:
            yield content

    async def _iterate(contents, step_notifier):
        return _iterate_impl(contents, step_notifier)

    source_iterator.iterate = _iterate

    source_selector = Mock()
    source_selector.select = AsyncMock(
        return_value=SourceSelectionResult(
            should_select=True,
            reason="Relevant",
            notification_message="Selected",
        )
    )

    reporting_agent = Mock()
    reporting_agent.generate = AsyncMock()
    reporting_agent.get_reports.return_value = _make_report_components()

    source_registry = Mock()
    memory_service = Mock()

    progress_notifier = Mock()
    progress_notifier.update = AsyncMock()

    chat_service = Mock()

    # Create orchestrator
    orchestrator = SWOTOrchestrator(
        step_notifier=step_notifier,
        source_collector=source_collector,
        source_selector=source_selector,
        source_iterator=source_iterator,
        reporting_agent=reporting_agent,
        source_registry=source_registry,
        progress_notifier=progress_notifier,
        memory_service=memory_service,
        chat_service=chat_service,
    )

    # Run orchestration
    plan = _make_plan()
    result = await orchestrator.run(company_name="ACME Corp", plan=plan)

    # Verify workflow
    source_collector.collect.assert_awaited_once()
    assert source_selector.select.await_count == 2  # Called for each content
    assert reporting_agent.generate.await_count == 2  # Called for each selected content
    reporting_agent.get_reports.assert_called_once()

    # Verify result
    assert isinstance(result, SWOTReportComponents)
    assert len(result.strengths) == 1


@pytest.mark.asyncio
async def test_orchestrator_skips_unselected_sources():
    """Test that sources marked as should_select=False are skipped."""
    step_notifier = Mock()
    step_notifier.notify = AsyncMock()

    contents = [_make_content("content_1", "Doc1"), _make_content("content_2", "Doc2")]

    source_collector = Mock()
    source_collector.collect = AsyncMock(return_value=contents)

    source_iterator = Mock()

    async def _iterate_impl(contents, step_notifier):
        for content in contents:
            yield content

    async def _iterate(contents, step_notifier):
        return _iterate_impl(contents, step_notifier)

    source_iterator.iterate = _iterate

    # First source selected, second skipped
    selection_results = [
        SourceSelectionResult(
            should_select=True,
            reason="Relevant",
            notification_message="Selected",
        ),
        SourceSelectionResult(
            should_select=False,
            reason="Not relevant",
            notification_message="Skipped",
        ),
    ]

    source_selector = Mock()
    source_selector.select = AsyncMock(side_effect=selection_results)

    reporting_agent = Mock()
    reporting_agent.generate = AsyncMock()
    reporting_agent.get_reports.return_value = _make_report_components()

    source_registry = Mock()
    memory_service = Mock()

    progress_notifier = Mock()
    progress_notifier.update = AsyncMock()

    chat_service = Mock()

    orchestrator = SWOTOrchestrator(
        step_notifier=step_notifier,
        source_collector=source_collector,
        source_selector=source_selector,
        source_iterator=source_iterator,
        reporting_agent=reporting_agent,
        source_registry=source_registry,
        progress_notifier=progress_notifier,
        memory_service=memory_service,
        chat_service=chat_service,
    )

    plan = _make_plan()
    await orchestrator.run(company_name="ACME Corp", plan=plan)

    # Only one source should be processed
    assert reporting_agent.generate.await_count == 1


@pytest.mark.asyncio
async def test_orchestrator_handles_empty_sources():
    """Test that orchestrator raises ValueError when there are no sources to process."""
    step_notifier = Mock()
    step_notifier.notify = AsyncMock()

    source_collector = Mock()
    source_collector.collect = AsyncMock(return_value=[])

    source_iterator = Mock()

    async def _iterate_impl(contents, step_notifier):
        for content in contents:
            yield content

    async def _iterate(contents, step_notifier):
        return _iterate_impl(contents, step_notifier)

    source_iterator.iterate = _iterate

    source_selector = Mock()
    reporting_agent = Mock()
    reporting_agent.get_reports.return_value = SWOTReportComponents(
        strengths=[], weaknesses=[], opportunities=[], threats=[]
    )

    source_registry = Mock()
    memory_service = Mock()

    progress_notifier = Mock()
    progress_notifier.update = AsyncMock()

    chat_service = Mock()

    orchestrator = SWOTOrchestrator(
        step_notifier=step_notifier,
        source_collector=source_collector,
        source_selector=source_selector,
        source_iterator=source_iterator,
        reporting_agent=reporting_agent,
        source_registry=source_registry,
        progress_notifier=progress_notifier,
        memory_service=memory_service,
        chat_service=chat_service,
    )

    plan = _make_plan()

    # Should raise ValueError when there are no sources
    with pytest.raises(ValueError, match="No sources to process"):
        await orchestrator.run(company_name="ACME Corp", plan=plan)

    # Verify collection was attempted
    source_collector.collect.assert_awaited_once()


@pytest.mark.asyncio
async def test_orchestrator_passes_correct_parameters():
    """Test that orchestrator passes correct parameters to each component."""
    step_notifier = Mock()
    step_notifier.notify = AsyncMock()

    content = _make_content()
    source_collector = Mock()
    source_collector.collect = AsyncMock(return_value=[content])

    source_iterator = Mock()

    async def _iterate_impl(contents, step_notifier):
        for content in contents:
            yield content

    async def _iterate(contents, step_notifier):
        return _iterate_impl(contents, step_notifier)

    source_iterator.iterate = _iterate

    source_selector = Mock()
    source_selector.select = AsyncMock(
        return_value=SourceSelectionResult(
            should_select=True,
            reason="Relevant",
            notification_message="Selected",
        )
    )

    reporting_agent = Mock()
    reporting_agent.generate = AsyncMock()
    reporting_agent.get_reports.return_value = _make_report_components()

    source_registry = Mock()
    memory_service = Mock()

    progress_notifier = Mock()
    progress_notifier.update = AsyncMock()

    chat_service = Mock()

    orchestrator = SWOTOrchestrator(
        step_notifier=step_notifier,
        source_collector=source_collector,
        source_selector=source_selector,
        source_iterator=source_iterator,
        reporting_agent=reporting_agent,
        source_registry=source_registry,
        progress_notifier=progress_notifier,
        memory_service=memory_service,
        chat_service=chat_service,
    )

    company_name = "ACME Corp"
    plan = _make_plan()
    await orchestrator.run(company_name=company_name, plan=plan)

    # Verify parameters passed to selector
    source_selector.select.assert_awaited_once()
    call_kwargs = source_selector.select.call_args.kwargs
    assert call_kwargs["company_name"] == company_name
    assert call_kwargs["content"] == content
    assert call_kwargs["step_notifier"] == step_notifier

    # Verify parameters passed to reporting agent
    reporting_agent.generate.assert_awaited_once()
    gen_kwargs = reporting_agent.generate.call_args.kwargs
    assert gen_kwargs["plan"] == plan
    assert gen_kwargs["content"] == content
    assert gen_kwargs["step_notifier"] == step_notifier
    assert gen_kwargs["source_registry"] == source_registry


@pytest.mark.asyncio
async def test_orchestrator_with_multiple_components():
    """Test orchestrator with a plan requesting multiple SWOT components."""
    step_notifier = Mock()
    step_notifier.notify = AsyncMock()

    content = _make_content()
    source_collector = Mock()
    source_collector.collect = AsyncMock(return_value=[content])

    source_iterator = Mock()

    async def _iterate_impl(contents, step_notifier):
        for content in contents:
            yield content

    async def _iterate(contents, step_notifier):
        return _iterate_impl(contents, step_notifier)

    source_iterator.iterate = _iterate

    source_selector = Mock()
    source_selector.select = AsyncMock(
        return_value=SourceSelectionResult(
            should_select=True,
            reason="Relevant",
            notification_message="Selected",
        )
    )

    reporting_agent = Mock()
    reporting_agent.generate = AsyncMock()
    reporting_agent.get_reports.return_value = SWOTReportComponents(
        strengths=[
            SWOTReportComponentSection(
                h2="Strength",
                entries=[SWOTReportSectionEntry(preview="P", content="C")],
            )
        ],
        weaknesses=[
            SWOTReportComponentSection(
                h2="Weakness",
                entries=[SWOTReportSectionEntry(preview="P", content="C")],
            )
        ],
        opportunities=[],
        threats=[],
    )

    source_registry = Mock()
    memory_service = Mock()

    progress_notifier = Mock()
    progress_notifier.update = AsyncMock()

    chat_service = Mock()

    orchestrator = SWOTOrchestrator(
        step_notifier=step_notifier,
        source_collector=source_collector,
        source_selector=source_selector,
        source_iterator=source_iterator,
        reporting_agent=reporting_agent,
        source_registry=source_registry,
        progress_notifier=progress_notifier,
        memory_service=memory_service,
        chat_service=chat_service,
    )

    # Plan with multiple components
    plan = SWOTPlan(
        objective="Full analysis",
        strengths=SWOTStepPlan(
            operation=SWOTOperation.GENERATE, modify_instruction=None
        ),
        weaknesses=SWOTStepPlan(
            operation=SWOTOperation.GENERATE, modify_instruction=None
        ),
        opportunities=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
        threats=SWOTStepPlan(
            operation=SWOTOperation.NOT_REQUESTED, modify_instruction=None
        ),
    )

    result = await orchestrator.run(company_name="ACME Corp", plan=plan)

    assert len(result.strengths) == 1
    assert len(result.weaknesses) == 1
    assert len(result.opportunities) == 0
    assert len(result.threats) == 0



"""Tests for the SWOTReportRegistry."""

import pytest

from unique_swot.services.generation.context import SWOTComponent
from unique_swot.services.generation.models.base import (
    SWOTReportComponentSection,
    SWOTReportSectionEntry,
)
from unique_swot.services.generation.models.registry import SWOTReportRegistry


def _make_section(h2="Test Section", num_entries=2):
    """Helper to create a test section."""
    entries = [
        SWOTReportSectionEntry(
            preview=f"Preview {i}", content=f"Content {i} [chunk_{i}]"
        )
        for i in range(num_entries)
    ]
    return SWOTReportComponentSection(h2=h2, entries=entries)


def test_registry_register_section():
    """Test registering a section returns a unique ID."""
    registry = SWOTReportRegistry()
    section = _make_section()

    section_id = registry.register_section(SWOTComponent.STRENGTHS, section)

    assert section_id.startswith("section_Strengths")


def test_registry_retrieve_section():
    """Test retrieving a registered section by ID."""
    registry = SWOTReportRegistry()
    section = _make_section("Test Strength")

    section_id = registry.register_section(SWOTComponent.STRENGTHS, section)
    retrieved = registry.retrieve_section(section_id)

    assert retrieved == section
    assert retrieved.h2 == "Test Strength"


def test_registry_retrieve_nonexistent_section():
    """Test retrieving a non-existent section returns None."""
    registry = SWOTReportRegistry()

    result = registry.retrieve_section("nonexistent_id")

    assert result is None


def test_registry_retrieve_component_sections():
    """Test retrieving all sections for a component."""
    registry = SWOTReportRegistry()

    section1 = _make_section("Strength 1")
    section2 = _make_section("Strength 2")

    registry.register_section(SWOTComponent.STRENGTHS, section1)
    registry.register_section(SWOTComponent.STRENGTHS, section2)
    registry.register_section(SWOTComponent.WEAKNESSES, _make_section("Weakness 1"))

    strengths = registry.retrieve_component_sections(SWOTComponent.STRENGTHS)

    assert len(strengths) == 2
    assert section1 in strengths
    assert section2 in strengths


def test_registry_retrieve_sections_for_component_json():
    """Test retrieving sections as JSON string."""
    registry = SWOTReportRegistry()

    section = _make_section("Test")
    section_id = registry.register_section(SWOTComponent.STRENGTHS, section)

    json_str = registry.retrieve_sections_for_component(SWOTComponent.STRENGTHS)

    assert section_id in json_str
    assert "Test" in json_str
    assert isinstance(json_str, str)


def test_registry_update_section():
    """Test updating an existing section."""
    registry = SWOTReportRegistry()

    original_section = _make_section("Original")
    section_id = registry.register_section(SWOTComponent.STRENGTHS, original_section)

    updated_section = _make_section("Updated")
    registry.update_section(section_id, updated_section)

    retrieved = registry.retrieve_section(section_id)
    assert retrieved.h2 == "Updated"


def test_registry_update_nonexistent_section_raises():
    """Test updating a non-existent section raises KeyError."""
    registry = SWOTReportRegistry()

    section = _make_section()

    with pytest.raises(KeyError):
        registry.update_section("nonexistent_id", section)


def test_registry_multiple_components():
    """Test registry handles multiple components independently."""
    registry = SWOTReportRegistry()

    registry.register_section(SWOTComponent.STRENGTHS, _make_section("S1"))
    registry.register_section(SWOTComponent.STRENGTHS, _make_section("S2"))
    registry.register_section(SWOTComponent.WEAKNESSES, _make_section("W1"))
    registry.register_section(SWOTComponent.OPPORTUNITIES, _make_section("O1"))
    registry.register_section(SWOTComponent.OPPORTUNITIES, _make_section("O2"))
    registry.register_section(SWOTComponent.OPPORTUNITIES, _make_section("O3"))
    registry.register_section(SWOTComponent.THREATS, _make_section("T1"))

    assert len(registry.retrieve_component_sections(SWOTComponent.STRENGTHS)) == 2
    assert len(registry.retrieve_component_sections(SWOTComponent.WEAKNESSES)) == 1
    assert len(registry.retrieve_component_sections(SWOTComponent.OPPORTUNITIES)) == 3
    assert len(registry.retrieve_component_sections(SWOTComponent.THREATS)) == 1


def test_registry_empty_component():
    """Test retrieving sections from an empty component."""
    registry = SWOTReportRegistry()

    sections = registry.retrieve_component_sections(SWOTComponent.STRENGTHS)

    assert len(sections) == 0
    assert sections == []


def test_registry_section_id_uniqueness():
    """Test that section IDs are unique."""
    registry = SWOTReportRegistry()

    section = _make_section()
    id1 = registry.register_section(SWOTComponent.STRENGTHS, section)
    id2 = registry.register_section(SWOTComponent.STRENGTHS, section)
    id3 = registry.register_section(SWOTComponent.WEAKNESSES, section)

    assert id1 != id2
    assert id1 != id3
    assert id2 != id3


def test_registry_retrieve_sections_exclude_items():
    """Test retrieving sections with exclude_items parameter."""
    registry = SWOTReportRegistry()

    section = _make_section("Test", num_entries=3)
    registry.register_section(SWOTComponent.STRENGTHS, section)

    # With exclude_items=True (default)
    json_with_exclusion = registry.retrieve_sections_for_component(
        SWOTComponent.STRENGTHS, exclude_items=True
    )

    # With exclude_items=False
    json_without_exclusion = registry.retrieve_sections_for_component(
        SWOTComponent.STRENGTHS, exclude_items=False
    )

    # Both should be valid JSON strings
    assert isinstance(json_with_exclusion, str)
    assert isinstance(json_without_exclusion, str)


def test_registry_preserves_section_data():
    """Test that registry preserves all section data."""
    registry = SWOTReportRegistry()

    entries = [
        SWOTReportSectionEntry(
            preview="Strong brand",
            content="The company has strong brand recognition [chunk_a]",
        ),
        SWOTReportSectionEntry(
            preview="Market leader", content="Leading position in the market [chunk_b]"
        ),
    ]
    section = SWOTReportComponentSection(h2="Market Position", entries=entries)

    section_id = registry.register_section(SWOTComponent.STRENGTHS, section)
    retrieved = registry.retrieve_section(section_id)

    assert retrieved.h2 == "Market Position"
    assert len(retrieved.entries) == 2
    assert retrieved.entries[0].preview == "Strong brand"
    assert retrieved.entries[1].content == "Leading position in the market [chunk_b]"



from pydantic import Field

from unique_swot.utils import StructuredOutputResult, StructuredOutputWithNotification


## SWOT Extraction Models
class SWOTExtractionFactsList(StructuredOutputWithNotification):
    facts: list[str] = Field(
        description="The list of facts extracted from the sources chunks."
    )


## SWOT Report Models
class SWOTReportSectionEntry(StructuredOutputResult):
    preview: str = Field(
        description="The preview of the section entry. Should be a concise preview that captures the main idea of the entry."
    )
    content: str = Field(description="The content of the section entry.")


class SWOTReportComponentSection(StructuredOutputResult):
    h2: str = Field(
        description="The subheading of the cluster. Should be concise sentence that summarizes the content of the entries in the cluster."
    )
    entries: list[SWOTReportSectionEntry] = Field(
        description="The list of entries in the section. Each entry should be a concise title and content that summarizes the content of the entry in the section."
    )


class SWOTReportComponents(StructuredOutputResult):
    strengths: list[SWOTReportComponentSection] = Field(
        description="The sections of the report. Each section should be a concise title and content that summarizes the content of the section."
    )
    weaknesses: list[SWOTReportComponentSection] = Field(
        description="The sections of the report. Each section should be a concise title and content that summarizes the content of the section."
    )
    opportunities: list[SWOTReportComponentSection] = Field(
        description="The sections of the report. Each section should be a concise title and content that summarizes the content of the section."
    )
    threats: list[SWOTReportComponentSection] = Field(
        description="The sections of the report. Each section should be a concise title and content that summarizes the content of the section."
    )

    def is_empty(self) -> bool:
        return all(len(section) == 0 for section in self.model_dump().values())



from logging import getLogger
from pathlib import Path
from typing import TypeVar
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field
from unique_toolkit import LanguageModelService
from unique_toolkit._common.validators import LMI
from unique_toolkit.content import Content, ContentChunk, ContentReference
from unique_toolkit.language_model.builder import MessagesBuilder

_LOGGER = getLogger(__name__)
_MAX_RETRIES = 3
T = TypeVar("T", bound=BaseModel)


def generate_unique_id(prefix: str) -> str:
    return f"{prefix}_{uuid4().hex[:8]}"


def load_template(parent_dir: Path, template_name: str) -> str:
    """Load a template from a text file."""
    with open(parent_dir / template_name, "r") as file:
        return file.read().strip()


class StructuredOutputResult(BaseModel):
    """This class is responsible for the result of the structured output generation."""

    model_config = ConfigDict(extra="forbid")


class StructuredOutputWithNotification(StructuredOutputResult):
    """This class is responsible for the result of the structured output generation with a notification message."""

    model_config = ConfigDict(extra="forbid")

    notification_message: str = Field(
        description="A notification message to entertain the user with highlighting key findings or results."
    )


async def generate_structured_output(
    *,
    user_message: str,
    system_prompt: str,
    llm: LMI,
    output_model: type[T],
    llm_service: LanguageModelService,
) -> T | None:
    """
    Call the LLM to produce a structured Pydantic model with retries.

    Returns None when every attempt fails.
    """
    _LOGGER.info(f"Generating structured output with {output_model.__name__} model")

    def _build_messages(error: str | None = None):
        builder = (
            MessagesBuilder()
            .system_message_append(system_prompt)
            .user_message_append(user_message)
        )
        if error:
            builder.user_message_append(
                f"The following error occurred, please retry and fix the error: \n ```{error}```"
            )
        return builder.build()

    last_error = ""
    for attempt in range(1, _MAX_RETRIES + 1):
        try:
            response = await llm_service.complete_async(
                model_name=llm.name,
                messages=_build_messages(last_error),
                structured_output_model=output_model,
                structured_output_enforce_schema=True,
            )
            return output_model.model_validate(response.choices[0].message.parsed)
        except Exception as exc:
            last_error = str(exc)
            _LOGGER.exception(
                "Error generating structured output (attempt %s/%s): %s",
                attempt,
                _MAX_RETRIES,
                last_error,
            )

    _LOGGER.error(
        "Failed to generate structured output after %s retries: %s",
        _MAX_RETRIES,
        last_error,
    )
    return None


def convert_content_chunk_to_reference(
    *,
    content_or_chunk: Content | ContentChunk,
    sequence_number: int | None = None,
) -> ContentReference:
    title = get_content_chunk_title(content_or_chunk)

    return ContentReference(
        url=f"unique://content/{content_or_chunk.id}",
        source_id=content_or_chunk.id,
        name=title,
        sequence_number=sequence_number or 0,
        source="SWOT-TOOL",
    )


def get_content_chunk_title(content_or_chunk: Content | ContentChunk) -> str:
    return content_or_chunk.title or content_or_chunk.key or "Unknown Title"



from typing import Annotated

from pydantic import BaseModel, Field
from unique_toolkit._common.pydantic.rjsf_tags import RJSFMetaTag
from unique_toolkit._common.pydantic_helpers import get_configuration_dict

from unique_swot.services.generation.context import SWOTComponent

_COMPONENT_DEFINITIONS: dict[SWOTComponent, str] = {
    SWOTComponent.STRENGTHS: (
        "Internal advantages, assets, capabilities, or resources that give "
        "the company an edge (e.g., brand, tech, talent, cash, IP, loyal customers)."
    ),
    SWOTComponent.WEAKNESSES: (
        "Internal limitations or gaps that hinder performance or competitiveness "
        "(e.g., execution issues, product gaps, debt, churn, reliance on single supplier)."
    ),
    SWOTComponent.OPPORTUNITIES: (
        "External trends or openings the company can capture for growth or advantage "
        "(e.g., market expansion, new tech adoption, regulatory tailwinds, partner ecosystems)."
    ),
    SWOTComponent.THREATS: (
        "External risks that could harm the company’s position or performance "
        "(e.g., competitors, substitutes, regulatory or macro headwinds, supply shocks)."
    ),
}


class ComponentDefinitionPromptConfig(BaseModel):
    model_config = get_configuration_dict()

    strengths_definition: Annotated[
        str,
        RJSFMetaTag.StringWidget.textarea(
            rows=len(_COMPONENT_DEFINITIONS[SWOTComponent.STRENGTHS].split("\n"))
        ),
    ] = Field(
        default=_COMPONENT_DEFINITIONS[SWOTComponent.STRENGTHS],
        description="The definition of the strengths component.",
    )
    weaknesses_definition: Annotated[
        str,
        RJSFMetaTag.StringWidget.textarea(
            rows=len(_COMPONENT_DEFINITIONS[SWOTComponent.WEAKNESSES].split("\n"))
        ),
    ] = Field(
        default=_COMPONENT_DEFINITIONS[SWOTComponent.WEAKNESSES],
        description="The definition of the weaknesses component.",
    )
    opportunities_definition: Annotated[
        str,
        RJSFMetaTag.StringWidget.textarea(
            rows=len(_COMPONENT_DEFINITIONS[SWOTComponent.OPPORTUNITIES].split("\n"))
        ),
    ] = Field(
        default=_COMPONENT_DEFINITIONS[SWOTComponent.OPPORTUNITIES],
        description="The definition of the opportunities component.",
    )
    threats_definition: Annotated[
        str,
        RJSFMetaTag.StringWidget.textarea(
            rows=len(_COMPONENT_DEFINITIONS[SWOTComponent.THREATS].split("\n"))
        ),
    ] = Field(
        default=_COMPONENT_DEFINITIONS[SWOTComponent.THREATS],
        description="The definition of the threats component.",
    )


def get_component_definition(
    component: SWOTComponent, config: ComponentDefinitionPromptConfig
) -> str:
    match component:
        case SWOTComponent.STRENGTHS:
            return config.strengths_definition
        case SWOTComponent.WEAKNESSES:
            return config.weaknesses_definition
        case SWOTComponent.OPPORTUNITIES:
            return config.opportunities_definition
        case SWOTComponent.THREATS:
            return config.threats_definition
        case _:
            raise ValueError(f"Unknown SWOT component: {component}")



"""Tests for the ReportRendererConfig."""

from unique_swot.services.report.config import RendererType, ReportRendererConfig


def test_report_renderer_config_ingest_docx_report_default():
    """Test that ingest_docx_report has a default value of False."""
    config = ReportRendererConfig()

    # Verify the default value
    assert config.ingest_docx_report is False
    assert isinstance(config.ingest_docx_report, bool)


def test_report_renderer_config_ingest_docx_report_can_be_set():
    """Test that ingest_docx_report can be explicitly set to True."""
    config = ReportRendererConfig(ingest_docx_report=True)

    # Verify the value is set correctly
    assert config.ingest_docx_report is True


def test_report_renderer_config_has_field_description():
    """Test that ingest_docx_report field has a description."""
    # Get the field info from the model
    field_info = ReportRendererConfig.model_fields["ingest_docx_report"]

    # Verify the description exists and contains expected content
    assert field_info.description is not None
    assert isinstance(field_info.description, str)
    assert len(field_info.description) > 0
    # The description should mention follow-up questions and side view
    assert (
        "follow-up" in field_info.description.lower()
        or "side view" in field_info.description.lower()
    )


def test_report_renderer_config_has_required_fields():
    """Test that ReportRendererConfig has all required fields."""
    config = ReportRendererConfig()

    # Verify required fields exist
    assert hasattr(config, "renderer_type")
    assert hasattr(config, "report_structure_template")
    assert hasattr(config, "report_body_template")
    assert hasattr(config, "docx_renderer_config")
    assert hasattr(config, "ingest_docx_report")


def test_report_renderer_config_renderer_type_default():
    """Test that renderer_type has a default value."""
    config = ReportRendererConfig()

    # Verify renderer_type is set to DOCX by default
    assert config.renderer_type == RendererType.DOCX


def test_report_renderer_config_templates_are_strings():
    """Test that template fields are strings."""
    config = ReportRendererConfig()

    # Verify templates are strings
    assert isinstance(config.report_structure_template, str)
    assert isinstance(config.report_body_template, str)



from typing import Callable, Generator, TypeVar

from tiktoken import get_encoding

T = TypeVar("T")


def batch_sequence_generator(
    *,
    encoder_name: str,
    source_batches: list[T],
    max_tokens_per_extraction_batch: int,
    serializer: Callable[[T], str],
) -> Generator[list[T], None, None]:
    encoder = get_encoding(encoder_name)
    current_batch = []
    current_tokens = 0
    for batch in source_batches:
        batch_tokens = encoder.encode(serializer(batch))
        if current_tokens + len(batch_tokens) > max_tokens_per_extraction_batch:
            yield current_batch
            current_batch = []
            current_tokens = 0
        current_batch.append(batch)
        current_tokens += len(batch_tokens)
    yield current_batch


def create_batch_notification(
    *,
    component: str,
    batch_index: int | None = None,
    total_batches: int,
) -> str:
    """
    Create a description message for batch processing notifications.

    Args:
        component: The SWOT component being processed
        batch_index: Current batch number (1-indexed), None for initial notification
        total_batches: Total number of batches

    Returns:
        Formatted description string
    """
    if total_batches == 1:
        return f"Processing {component}..."

    if batch_index is None:
        return f"This document is too large to extract. It will be split into {total_batches} batches."

    return f"Processing {component} (Batch {batch_index} of {total_batches})..."



from pydantic import BaseModel, Field
from unique_toolkit._common.pydantic_helpers import get_configuration_dict

from unique_swot.services.generation.agentic.prompts.commands.config import (
    CommandsPromptConfig,
)
from unique_swot.services.generation.agentic.prompts.definition.config import (
    ComponentDefinitionPromptConfig,
)
from unique_swot.services.generation.agentic.prompts.extraction.config import (
    ExtractionPromptConfig,
)
from unique_swot.services.generation.agentic.prompts.plan.config import PlanPromptConfig


class AgenticPromptsConfig(BaseModel):
    model_config = get_configuration_dict()

    commands_prompt_config: CommandsPromptConfig = Field(
        default=CommandsPromptConfig(),
        description="The prompt config for the commands.",
    )

    extraction_prompt_config: ExtractionPromptConfig = Field(
        default=ExtractionPromptConfig(),
        description="The prompt config for the extraction.",
    )

    plan_prompt_config: PlanPromptConfig = Field(
        default=PlanPromptConfig(),
        description="The prompt config for the plan.",
    )

    definition_prompt_config: ComponentDefinitionPromptConfig = Field(
        default=ComponentDefinitionPromptConfig(),
        description="The prompt config for the definition.",
    )



import asyncio
from datetime import datetime
from logging import getLogger
from typing import Sequence

from unique_quartr.constants.document_types import DocumentType
from unique_quartr.constants.event_types import EventType
from unique_quartr.endpoints.schemas import EventDto
from unique_quartr.service import DocumentDto, QuartrService
from unique_quartr.transcripts.earnings_call import QuartrEarningsCallTranscript
from unique_toolkit import KnowledgeBaseService
from unique_toolkit._common.docx_generator import DocxGeneratorService
from unique_toolkit.content import Content

from unique_swot.services.session.schema import UniqueCompanyListing

_LOGGER = getLogger(__name__)

_MAX_INGESTION_RETRIES = 100
_INGESTION_RETRY_DELAY = 1


async def collect_earnings_calls(
    *,
    docx_generator_service: DocxGeneratorService,
    knowledge_base_service: KnowledgeBaseService,
    quartr_service: QuartrService,
    upload_scope_id: str,
    company: UniqueCompanyListing,
    earnings_call_start_date: datetime,
) -> list[Content]:
    event_type_ids = quartr_service.get_event_subtype_ids_from_event_types(
        [EventType.EARNINGS_CALL]
    )
    document_ids = quartr_service.get_document_ids_from_document_types(
        [DocumentType.TRANSCRIPT, DocumentType.IN_HOUSE_TRANSCRIPT]
    )

    events = quartr_service.fetch_company_events(
        company_ids=[company.id],
        event_ids=event_type_ids,
        start_date=earnings_call_start_date.strftime("%Y-%m-%d"),
    )

    events_mapping = {int(event.id): event for event in events.data}

    documents = quartr_service.fetch_event_documents(
        event_ids=list(events_mapping.keys()),
        document_ids=document_ids,
    )

    available_contents_in_knowledge_base: list[Content] = []
    list_of_transcripts_to_ingest: list[DocumentDto] = []

    document_collection = {}
    for document in documents.data:
        if document.event_id is None:
            _LOGGER.warning(
                f"Document {document.id} has no event id. This is an unexpected behaviour. Skipping this document."
            )
            continue

        event = events_mapping[int(document.event_id)]
        document_title = f"{company.name} {event.title} Earnings Call"

        # Filtering step to remove duplicates (Happens when there are both In-House and External Transcripts for the same event)
        if document_title in document_collection:
            # If current is In-House Transcript, override the existing document as it is more complete
            if document.type_id == DocumentType.IN_HOUSE_TRANSCRIPT:
                document_collection[document_title] = document
        else:
            document_collection[document_title] = document

    for document_title, document in document_collection.items():
        # Try to get the content from the knowledge_base if already exists
        content = _try_getting_content_from_knowledge_base(
            knowledge_base_service=knowledge_base_service,
            document_title=document_title,
            scope_id=upload_scope_id,
        )
        if content is not None:
            _LOGGER.info(
                f"Content {document_title} already exists in the knowledge base"
            )
            available_contents_in_knowledge_base.append(content)
        else:
            _LOGGER.info(
                f"Content {document_title} not found in the knowledge base and will be added to ingestion queue!"
            )
            list_of_transcripts_to_ingest.append(document)

    ingested_contents = await _ingest_all_transcripts(
        list_of_transcripts_to_ingest=list_of_transcripts_to_ingest,
        knowledge_base_service=knowledge_base_service,
        docx_generator_service=docx_generator_service,
        upload_scope_id=upload_scope_id,
        company=company,
        events_mapping=events_mapping,
    )

    # Merging the available contents in the knowledge base and the ingested contents
    all_contents = available_contents_in_knowledge_base + ingested_contents

    return all_contents


async def _ingest_all_transcripts(
    *,
    list_of_transcripts_to_ingest: list[DocumentDto],
    knowledge_base_service: KnowledgeBaseService,
    docx_generator_service: DocxGeneratorService,
    upload_scope_id: str,
    company: UniqueCompanyListing,
    events_mapping: dict[int, EventDto],
) -> list[Content]:
    tasks = []

    for document in list_of_transcripts_to_ingest:
        if document.event_id is None:
            _LOGGER.warning(
                f"Document {document.id} has no event id. This is an unexpected behaviour. Skipping this document."
            )
            continue

        event = events_mapping[int(document.event_id)]
        document_title = f"{company.name} {event.title} Earnings Call"

        task = asyncio.create_task(
            _ingest_transcript_and_get_content(
                knowledge_base_service=knowledge_base_service,
                docx_generator_service=docx_generator_service,
                scope_id=upload_scope_id,
                document_title=document_title,
                quarter_transcript_url=document.file_url,
                quarter_event_date=event.date,
            )
        )
        tasks.append(task)

    ingested_contents: Sequence[Content | None] = await asyncio.gather(*tasks)
    return [content for content in ingested_contents if content is not None]


def _try_getting_content_from_knowledge_base(
    *,
    knowledge_base_service: KnowledgeBaseService,
    document_title: str,
    scope_id: str,
) -> Content | None:
    where_clause = {
        "AND": [
            {"title": {"contains": document_title}},
            {"ownerId": {"equals": scope_id}},
        ]
    }
    contents = knowledge_base_service.search_contents(where=where_clause)
    if len(contents) == 0:
        return None

    assert len(contents) == 1, (
        "Expected exactly one content to be found for the given document title"
    )

    return contents[0]


async def _ingest_transcript_and_get_content(
    *,
    knowledge_base_service: KnowledgeBaseService,
    docx_generator_service: DocxGeneratorService,
    scope_id: str,
    document_title: str,
    quarter_transcript_url: str,
    quarter_event_date: datetime,
) -> Content | None:
    # Loading the transcript from the quarter transcript url
    transcript = QuartrEarningsCallTranscript.from_quartr_transcript_url(
        quarter_transcript_url
    )

    # Converting the transcript to markdown
    markdown = transcript.to_markdown()

    # Parsing the markdown to a list of content fields
    docx_bytes = docx_generator_service.parse_markdown_to_list_content_fields(markdown)

    # Generating the docx from the content fields and the template
    docx_bytes = docx_generator_service.generate_from_template(
        docx_bytes,
        {
            "title": document_title,
            "date": quarter_event_date.strftime("%Y-%m-%d"),
        },
    )

    # Early return if the docx generation failed
    if docx_bytes is None:
        _LOGGER.warning(f"Failed to generate the docx for the content {document_title}")
        return None

    # Uploading the docx to the knowledge base
    content = knowledge_base_service.upload_content_from_bytes(
        content=docx_bytes,
        content_name=document_title + ".docx",
        mime_type="application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        scope_id=scope_id,
    )

    # Waiting for the content to be ready
    content = await _wait_for_content_to_be_ready(content.id, knowledge_base_service)

    if content is None:
        _LOGGER.warning(f"Content {document_title} ingestion failed or still ingesting")
    else:
        _LOGGER.info(f"Content {document_title} ingestion finished")

    return content


async def _wait_for_content_to_be_ready(
    content_id: str, knowledge_base_service: KnowledgeBaseService
) -> Content | None:
    for _ in range(_MAX_INGESTION_RETRIES):
        contents = knowledge_base_service.search_contents(
            where={"id": {"equals": content_id}}, include_failed_content=True
        )
        if len(contents) > 0:
            ingestion_state = contents[0].ingestion_state
            _LOGGER.debug(f"Content {content_id} ingestion state: {ingestion_state}")
            if ingestion_state == "FINISHED":
                return contents[0]
            elif ingestion_state == "FAILED":
                raise ValueError(f"Content {content_id} ingestion failed")
            else:
                await asyncio.sleep(_INGESTION_RETRY_DELAY)

    return None



"""Tests for the SwotMemoryService."""

from unittest.mock import Mock

from pydantic import BaseModel

from unique_swot.services.memory.base import Memory, SwotMemoryService


class TestModel(BaseModel):
    """Test model for memory service."""

    value: str
    count: int


def test_memory_service_get_returns_none_when_no_memory():
    """Test that get returns None when no memory exists."""
    kb_service = Mock()
    stm_service = Mock()
    stm_service.find_latest_memory.return_value = None

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    result = service.get(TestModel)

    assert result is None


def test_memory_service_get_retrieves_from_kb():
    """Test that get retrieves content from knowledge base."""
    kb_service = Mock()
    kb_service.download_content_to_bytes.return_value = (
        b'{"value": "test", "count": 42}'
    )

    stm_service = Mock()
    memory_data = {"id": "content_123", "file_name": "TestModel_abc.json"}
    stm_service.find_latest_memory.return_value = Mock(data=memory_data)

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    result = service.get(TestModel)

    assert result is not None
    assert result.value == "test"
    assert result.count == 42
    kb_service.download_content_to_bytes.assert_called_once_with(
        content_id="content_123"
    )


def test_memory_service_get_handles_kb_error():
    """Test that get handles knowledge base errors gracefully."""
    kb_service = Mock()
    kb_service.download_content_to_bytes.side_effect = Exception("KB error")

    stm_service = Mock()
    memory_data = {"id": "content_123", "file_name": "TestModel_abc.json"}
    stm_service.find_latest_memory.return_value = Mock(data=memory_data)

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    result = service.get(TestModel)

    assert result is None


def test_memory_service_set_creates_new_memory():
    """Test that set creates new memory when none exists."""
    kb_service = Mock()
    kb_service.upload_content_from_bytes.return_value = Mock(id="new_content_123")

    stm_service = Mock()
    stm_service.find_latest_memory.return_value = None

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    model = TestModel(value="test", count=42)
    service.set(model)

    # Verify upload was called
    kb_service.upload_content_from_bytes.assert_called_once()
    call_kwargs = kb_service.upload_content_from_bytes.call_args.kwargs
    assert call_kwargs["scope_id"] == "test_scope"
    assert call_kwargs["skip_ingestion"] is True

    # Verify short-term memory was created
    stm_service.create_memory.assert_called_once()


def test_memory_service_set_updates_existing_memory():
    """Test that set updates existing memory."""
    kb_service = Mock()
    kb_service.upload_content_from_bytes.return_value = Mock(id="updated_content_123")

    stm_service = Mock()
    memory_data = {"id": "old_content_123", "file_name": "TestModel_existing.json"}
    stm_service.find_latest_memory.return_value = Mock(data=memory_data)

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    model = TestModel(value="updated", count=100)
    service.set(model)

    # Verify upload was called with existing file name
    call_kwargs = kb_service.upload_content_from_bytes.call_args.kwargs
    assert call_kwargs["content_name"] == "TestModel_existing.json"


def test_memory_service_set_requires_cache_scope_id():
    """Test that set raises error when cache_scope_id is not set."""
    kb_service = Mock()
    stm_service = Mock()
    stm_service.find_latest_memory.return_value = None

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="",  # Empty scope ID
    )

    model = TestModel(value="test", count=42)
    service.set(model)

    # Should not upload without scope ID
    kb_service.upload_content_from_bytes.assert_not_called()


def test_memory_service_set_handles_upload_error():
    """Test that set handles upload errors gracefully."""
    kb_service = Mock()
    kb_service.upload_content_from_bytes.side_effect = Exception("Upload error")

    stm_service = Mock()
    stm_service.find_latest_memory.return_value = None

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    model = TestModel(value="test", count=42)
    # Should not raise exception
    service.set(model)


def test_memory_service_uses_model_name_as_key():
    """Test that memory service uses model class name as key."""
    kb_service = Mock()
    kb_service.upload_content_from_bytes.return_value = Mock(id="content_123")

    stm_service = Mock()
    stm_service.find_latest_memory.return_value = None

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    model = TestModel(value="test", count=42)
    service.set(model)

    # Verify key is class name
    stm_service.create_memory.assert_called_once()
    call_args = stm_service.create_memory.call_args
    assert call_args[0][0] == "TestModel"


def test_memory_service_serializes_json_correctly():
    """Test that memory service serializes models to JSON correctly."""
    kb_service = Mock()
    kb_service.upload_content_from_bytes.return_value = Mock(id="content_123")

    stm_service = Mock()
    stm_service.find_latest_memory.return_value = None

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    model = TestModel(value="test", count=42)
    service.set(model)

    # Verify JSON serialization
    call_kwargs = kb_service.upload_content_from_bytes.call_args.kwargs
    content = call_kwargs["content"]
    assert b'"value": "test"' in content
    assert b'"count": 42' in content


def test_memory_model_validation():
    """Test Memory model validation."""
    memory = Memory(id="content_123", file_name="test.json")

    assert memory.id == "content_123"
    assert memory.file_name == "test.json"


def test_memory_service_find_latest_memory_error_handling():
    """Test that find_latest_memory errors are handled."""
    kb_service = Mock()
    stm_service = Mock()
    stm_service.find_latest_memory.side_effect = Exception("STM error")

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    result = service.get(TestModel)

    # Should return None on error
    assert result is None


def test_memory_service_mime_type_is_text_plain():
    """Test that uploaded content has text/plain mime type."""
    kb_service = Mock()
    kb_service.upload_content_from_bytes.return_value = Mock(id="content_123")

    stm_service = Mock()
    stm_service.find_latest_memory.return_value = None

    service = SwotMemoryService(
        short_term_memory_service=stm_service,
        knowledge_base_service=kb_service,
        cache_scope_id="test_scope",
    )

    model = TestModel(value="test", count=42)
    service.set(model)

    call_kwargs = kb_service.upload_content_from_bytes.call_args.kwargs
    assert call_kwargs["mime_type"] == "text/plain"



"""Tests for the SwotAnalysisTool service."""

from unittest.mock import AsyncMock, Mock

import pytest
from unique_toolkit.agentic.tools.schemas import ToolCallResponse
from unique_toolkit.language_model import LanguageModelFunction

from unique_swot.service import SwotAnalysisTool


def test_get_evaluation_checks_returns_empty_list():
    """Test that get_evaluation_checks_based_on_tool_response returns an empty list.

    This method is part of the Tool interface and indicates what evaluation
    metrics should be run based on a tool response. For SWOT Analysis, we
    don't require any specific evaluation checks, so it returns an empty list.
    """
    # Import the class to ensure coverage tracking
    from unique_swot.service import SwotAnalysisTool

    # Create an instance without full initialization using object.__new__
    # This avoids the complex initialization that requires event data
    tool = object.__new__(SwotAnalysisTool)

    # Create a mock tool response
    tool_response = Mock(spec=ToolCallResponse)
    tool_response.id = "test-id"
    tool_response.name = "SwotAnalysis"
    tool_response.content = "Test content"
    tool_response.content_chunks = []

    # Call the method - it doesn't use self attributes, just returns []
    result = tool.get_evaluation_checks_based_on_tool_response(tool_response)

    # Verify it returns an empty list
    assert result == []
    assert isinstance(result, list)
    assert len(result) == 0


def test_evaluation_check_list_returns_empty_list():
    """Test that evaluation_check_list returns an empty list.

    This method indicates the full list of evaluation metrics that this tool
    supports. For SWOT Analysis, no specific evaluation metrics are needed.
    """
    # Create an instance without full initialization
    tool = object.__new__(SwotAnalysisTool)

    # Call the method
    result = tool.evaluation_check_list()

    # Verify it returns an empty list
    assert result == []
    assert isinstance(result, list)


def test_takes_control_returns_true():
    """Test that takes_control returns True.

    The SWOT Analysis tool takes control of the conversation and doesn't
    want the orchestrator to intervene during execution.
    """
    # Create an instance without full initialization
    tool = object.__new__(SwotAnalysisTool)

    # Call the method
    result = tool.takes_control()

    # Verify it returns True
    assert result is True
    assert isinstance(result, bool)


@pytest.mark.asyncio
async def test_run_calls_set_completed_at_in_finally_on_exception():
    """Test that the finally block calls set_completed_at=True even when an exception occurs.

    This ensures that the chat message is always marked as completed, regardless of
    whether the SWOT analysis succeeds or fails.
    """
    # Create a tool instance
    tool = object.__new__(SwotAnalysisTool)

    # Mock the required attributes
    tool._chat_service = AsyncMock()
    tool._try_load_session_config = Mock()

    # Create a mock session config that will pass the initial check
    mock_session_config = Mock()
    mock_session_config.swot_analysis.company_listing.name = "Test Company"
    mock_session_config.swot_analysis.render_session_info = Mock(
        return_value="Test Info"
    )
    tool._try_load_session_config.return_value = mock_session_config

    # Mock _get_progress_notifier to return a mock that raises an exception
    mock_progress_notifier = AsyncMock()
    mock_progress_notifier.start = AsyncMock(side_effect=Exception("Test error"))
    mock_progress_notifier.finish = AsyncMock()
    tool._get_progress_notifier = Mock(return_value=mock_progress_notifier)

    # Create a mock tool call
    tool_call = Mock(spec=LanguageModelFunction)
    tool_call.id = "test-call-id"
    tool_call.arguments = {"operation": "generate"}

    # Run the method - it should catch the exception
    result = await tool.run(tool_call)

    # Verify the finally block was executed
    # The finally block should call modify_assistant_message_async with set_completed_at=True
    assert tool._chat_service.modify_assistant_message_async.call_count >= 1

    # Find the call with set_completed_at=True (from the finally block)
    calls = tool._chat_service.modify_assistant_message_async.call_args_list
    finally_call_found = False
    for call in calls:
        if call.kwargs.get("set_completed_at") is True and "content" not in call.kwargs:
            finally_call_found = True
            break

    assert finally_call_found, (
        "Finally block should call modify_assistant_message_async with set_completed_at=True"
    )

    # Verify the exception was handled and a ToolCallResponse was returned
    assert isinstance(result, ToolCallResponse)
    assert "Error running SWOT plan" in result.content


@pytest.mark.asyncio
async def test_run_returns_early_when_no_session_config():
    """Test that run returns early when session config is invalid.

    This tests the early return path where the finally block should NOT be executed
    because the try block is never entered.
    """
    # Create a tool instance
    tool = object.__new__(SwotAnalysisTool)

    # Mock the required attributes
    tool._chat_service = AsyncMock()
    tool._try_load_session_config = Mock(return_value=None)  # No session config
    tool.name = "SwotAnalysis"

    # Create a mock tool call
    tool_call = Mock(spec=LanguageModelFunction)
    tool_call.id = "test-call-id"

    # Run the method
    result = await tool.run(tool_call)

    # Verify it returned early with an error
    assert isinstance(result, ToolCallResponse)
    assert "not provided a valid configuration" in result.content

    # Verify modify_assistant_message_async was called with set_completed_at=True
    # (but this is in the early return, not the finally block)
    tool._chat_service.modify_assistant_message_async.assert_called_once()
    assert (
        tool._chat_service.modify_assistant_message_async.call_args.kwargs[
            "set_completed_at"
        ]
        is True
    )



from typing import Self

from pydantic import BaseModel
from unique_toolkit import ChatService
from unique_toolkit.chat.schemas import (
    MessageLogDetails,
    MessageLogStatus,
    MessageLogUncitedReferences,
)
from unique_toolkit.content.schemas import ContentReference


class MessageLogRegistryItem(BaseModel):
    title: str
    description: list[str]
    sources: list[ContentReference]
    message_log_id: str
    order: int
    progress: int
    completed: bool

    @classmethod
    async def create(
        cls,
        *,
        chat_service: ChatService,
        order: int,
        message_id: str,
        title: str,
        description: str,
        sources: list[ContentReference] = [],
        progress: int = 0,
        completed: bool = False,
    ) -> Self:
        text = f"**{title}** _{progress}%_\n - {description}"

        status = (
            MessageLogStatus.RUNNING if not completed else MessageLogStatus.COMPLETED
        )

        message_log = await chat_service.create_message_log_async(
            message_id=message_id,
            order=order,
            text=text,
            status=status,
            details=MessageLogDetails(data=[]),
            uncited_references=MessageLogUncitedReferences(data=sources),
            references=sources,
        )

        assert message_log.message_log_id is not None

        return cls(
            title=title,
            description=[description],
            sources=sources,
            message_log_id=message_log.message_log_id,
            order=order,
            progress=progress,
            completed=completed,
        )

    async def update(
        self,
        chat_service: ChatService,
        *,
        description: str,
        sources: list[ContentReference],
        progress: int | None = None,
        completed: bool = False,
    ) -> Self:
        self.description.append(description)
        if progress is not None:
            self.progress = progress
        text = (
            f"**{self.title}** _{self.progress}%_\n - {'\n - '.join(self.description)}"
        )
        status = (
            MessageLogStatus.RUNNING if not completed else MessageLogStatus.COMPLETED
        )
        self.sources.extend(sources)
        await chat_service.update_message_log_async(
            message_log_id=self.message_log_id,
            text=text,
            order=self.order,
            status=status,
            details=MessageLogDetails(data=[]),
            uncited_references=MessageLogUncitedReferences(data=self.sources),
            references=sources,
        )
        return self


class LogRegistry:
    def __init__(self):
        self._log_registry: dict[str, MessageLogRegistryItem] = {}
        self._total_number_of_references: int = 0

    async def add(
        self,
        chat_service: ChatService,
        message_id: str,
        title: str,
        description: str = "",
        sources: list[ContentReference] = [],
        progress: int | None = None,
        completed: bool = False,
    ):
        if sources:
            for index, source in enumerate(sources):
                source.sequence_number = index + self._total_number_of_references

        if title not in self._log_registry:
            if progress is None:
                progress = 0
            self._log_registry[title] = await MessageLogRegistryItem.create(
                chat_service=chat_service,
                order=len(self._log_registry) + 99,
                message_id=message_id,
                title=title,
                description=description,
                sources=sources,
                progress=progress,
                completed=completed,
            )
        else:
            await self._log_registry[title].update(
                chat_service=chat_service,
                description=description,
                sources=sources,
                progress=progress,
                completed=completed,
            )
        self._total_number_of_references += len(sources)
